import matplotlib
import matplotlib.cbook as mpcb
import matplotlib.pyplot as pyplot
import numpy as np
from ..config import mycfg
from ..config import ConfigDescriptor as cfgdec
import units
import models
from pyspeckit.specwarnings import warn
import interactive
import inspect
import copy

class Registry(object):
    """
    This class is a simple wrapper to prevent fitter properties from being globals
    """

    def __init__(self):
        self.npars = {}
        self.multifitters = {}
        self.singlefitters = {}
        self.fitkeys = {}
        self.associatedkeys = {}

        self._interactive_help_message_root = """

'?' will print this help message again. The / / keys are mnemonics.
1. Left-click or hit 'p' (/p/ick) with the cursor over the plot at both of the
two desired X-values to select a fitting range.  You can e/x/clude parts of the
spectrum by hitting 'x' at two positions.  
2. Then /m/iddle-click or hit 'm' twice to select (/m/ark) a peak and width -
the first mark should be on the peak of the line, the second should be at the
approximate half-max point on the curve.  
3. When you're done, right-click or hit 'd' to perform the fit and disconnect
the mouse and keyboard (/d/isconnect because you're /d/one).  Any time before
you're /d/one, you can select a different fitter (see below).

You can select different fitters to use with the interactive fitting routine.
The default is gaussian ('g'), all options are listed below:
        """
        self._make_interactive_help_message()

    def add_fitter(self, name, function, npars, multisingle='single',
        override=False, key=None):
        ''' 
        Register a fitter function.

        Required Arguments:

            *name*: [ string ]
                The fit function name. 

            *function*: [ function ]
                The fitter function.  Single-fitters should take npars + 1 input
                parameters, where the +1 is for a 0th order baseline fit.  They
                should accept an X-axis and data and standard fitting-function
                inputs (see, e.g., gaussfitter).  Multi-fitters should take N *
                npars, but should also operate on X-axis and data arguments.

            *npars*: [ int ]
                How many parameters does the function being fit accept?

        Optional Keyword Arguments:

            *multisingle*: [ 'multi' | 'single' ] 
                Is the function a single-function fitter (with a background), or
                does it allow N copies of the fitting function?

            *override*: [ True | False ]
                Whether to override any existing type if already present.

            *key*: [ char ]
                Key to select the fitter in interactive mode
        '''


        if multisingle == 'single':
            if not name in self.singlefitters or override:
                self.singlefitters[name] = function
        elif multisingle == 'multi':
            if not name in self.multifitters or override:
                self.multifitters[name] = function
        elif name in self.singlefitters or name in self.multifitters:
            raise Exception("Fitting function %s is already defined" % name)

        if key is not None:
            self.fitkeys[key] = name
            self._make_interactive_help_message()
        self.npars[name] = npars
        self.associated_keys = dict(zip(self.fitkeys.values(),self.fitkeys.keys()))

    def _make_interactive_help_message(self):
        """
        Generate the interactive help message from the fitkeys
        """
        self.interactive_help_message = (
                self._interactive_help_message_root + 
                "\n" +
                "\n".join(["'%s' - select fitter %s" % (key,name) for key,name in self.fitkeys.items()]) +
                "\n" # trailing \n so that users' input is on a fresh line
                )


default_Registry = Registry()
default_Registry.add_fitter('ammonia',models.ammonia_model(multisingle='multi'),6,multisingle='multi',key='a')
default_Registry.add_fitter('ammonia_tau',models.ammonia_model_vtau(multisingle='multi'),6,multisingle='multi')
# not implemented default_Registry.add_fitter(Registry,'ammonia',models.ammonia_model(multisingle='single'),6,multisingle='single',key='A')
default_Registry.add_fitter('formaldehyde',models.formaldehyde_fitter,3,multisingle='multi',key='F') # CAN'T USE f!  reserved for fitting
default_Registry.add_fitter('formaldehyde',models.formaldehyde_vheight_fitter,3,multisingle='single')
default_Registry.add_fitter('gaussian',models.gaussian_fitter(multisingle='multi'),3,multisingle='multi',key='g')
default_Registry.add_fitter('gaussian',models.gaussian_fitter(multisingle='single'),3,multisingle='single')
default_Registry.add_fitter('voigt',models.voigt_fitter(multisingle='multi'),4,multisingle='multi',key='v')
default_Registry.add_fitter('voigt',models.voigt_fitter(multisingle='single'),4,multisingle='single')
default_Registry.add_fitter('lorentzian',models.lorentzian_fitter(multisingle='multi'),3,multisingle='multi',key='L')
default_Registry.add_fitter('lorentzian',models.lorentzian_fitter(multisingle='single'),3,multisingle='single')
default_Registry.add_fitter('hill5',models.hill5infall.hill5_fitter,5,multisingle='multi')
default_Registry.add_fitter('hcn',models.hcn.hcn_vtau_fitter,4,multisingle='multi')


class Specfit(interactive.Interactive):

    def __init__(self, Spectrum, Registry=None):
        super(Specfit, self).__init__(Spectrum, interactive_help_message=Registry.interactive_help_message)
        self.model = None
        self.parinfo = None
        self.modelpars = None
        self.modelerrs = None
        self.modelplot = []
        self.modelcomponents = None
        self._plotted_components = []
        self.npeaks = 0
        #self.nclicks_b1 = 0
        #self.nclicks_b2 = 0
        #self.xmin = 0
        #self.xmax = Spectrum.data.shape[0]
        self.button2action = self.guesspeakwidth
        self.guesses = []
        self.click = 0
        self.fitkwargs = {}
        self.auto = False
        self.fitleg=None
        self.residuals=None
        self.setfitspec()
        self.fittype = 'gaussian'
        self.measurements = None
        self.vheight=False # vheight must be a boolean, can't be none
        self._component_kwargs = {}
        self.Registry = Registry
        self.autoannotate = mycfg['autoannotate']
        #self.seterrspec()
        
    @cfgdec
    def __call__(self, interactive=False, multifit=False, usemoments=True,
            clear_all_connections=True, debug=False, 
            guesses=None, save=True, fittype='gaussian', annotate=None,
            show_components=None, use_lmfit=False, verbose=True, clear=True,
            fit_plotted_area=True, use_window_limits=None,
            vheight=None, **kwargs):
        """
        Fit gaussians (or other model functions) to a spectrum

        Parameters
        ----------
        interactive : boolean
            The plotter window will go into interactive mode.  See
            self.interactive_help_message for details on how to use the
            interactive fitter.
        multifit : boolean
            If false, only a single peak is allower, but a "height" (0'th-order
            baseline) will be fit simultaneously with that peak.
        fittype : str
            The model to use.  Model must be registered in self.Registry.  
            gaussian, lorentzian, and voigt profiles are registered by default
        guesses : list
            A list of guesses.  Guesses must have length = n*number of parameters
            in model.  Guesses are *required* for multifit fits (there is no
            automated guessing for most models)
            EXAMPLE: for single-fit gaussian
            guesses = [height,amplitude,center,width]
        use_lmfit : boolean
            If lmfit-py (https://github.com/newville/lmfit-py) is installed, you
            can use it instead of the pure-python (but slow) mpfit.
        fit_plotted_area : boolean
            If no other limits are specified, the plotter's xmin/xmax will be
            used to define the fit region.  Only respects the x-axis limits,
            not the y-axis limits.
        use_window_limits : boolean
            If `fit_plotted_area==True` and no other limits are specified, will
            use the displayed window area (as set by the zoom tools) as the
            fitting range.  Only respects the x-axis limits, not the y-axis
            limits.


        Plotter-related Parameters
        --------------------------
        annotate : None or boolean
            If None, will use the default stored in self.annotate, otherwise
            overwrites.  Annotations will appear on the plot if a plotter
            exists.
        show_components : boolean
            Show the individual components of a multi-component fit (defaults
            to blue)
        clear : boolean
            Clear previous fitter plots before overplotting the fit?

        
        Advanced Parameters
        -------------------
        clear_all_connections : boolean
            Clear all of the interactive connections from a previous interactive
            session (e.g., a baseline fitting session) before continuing?
        usemoments : boolean
            Use the moments of the spectrum as input guesses.  Only works
            for gaussian and gaussian-like models.  Only works for single-fit
            mode (not multifit)
            DEPRECATED
        debug : boolean
            Print debug statements?
        save : boolean
            Save best fits in the FITS header as keywords?  ONLY IMPLEMENTED
            FOR GAUSSIANS
        verbose : boolean
            Print out extra stuff
        vheight : None or boolean
            if None, defaults to self.vheight, otherwise overrides
            Determines whether a 0th order baseline will be fit along with the
            line
        

        """

        if clear: self.clear()
        self.selectregion(verbose=verbose, debug=debug,
                fit_plotted_area=fit_plotted_area,
                use_window_limits=use_window_limits, **kwargs)
        for arg in ['xmin','xmax','xtype','reset']: 
            if arg in kwargs: kwargs.pop(arg)

        if self.fittype != fittype: self.fittype = fittype

        # multifit = True if the right number of guesses are passed
        if guesses is not None:
            if len(guesses) > 5:
                multifit = True

        self.npeaks = 0
        self.fitkwargs = kwargs
        if interactive:
            if self.Spectrum.plotter.axis is None:
                raise Exception("Interactive fitting requires a plotter.")
            # reset button count & guesses on every __call__
            self.nclicks_b1 = 0
            self.nclicks_b2 = 0
            self.guesses = []

            self.start_interactive(clear_all_connections=clear_all_connections, debug=debug, **kwargs)
        elif ((multifit or multifit is None) and self.fittype in self.Registry.multifitters) or guesses is not None:
            if guesses is None:
                print "You must input guesses when using multifit.  Also, baseline (continuum fit) first!"
                return
            else:
                self.guesses = guesses
                self.multifit(show_components=show_components, verbose=verbose,
                        debug=debug, use_lmfit=use_lmfit, annotate=annotate,
                        **kwargs)
        # SINGLEFITTERS SHOULD BE PHASED OUT
        elif self.fittype in self.Registry.singlefitters:
            #print "Non-interactive, 1D fit with automatic guessing"
            if (self.Spectrum.baseline.order is None and vheight is None) or vheight:
                self.Spectrum.baseline.order=0
                self.peakbgfit(usemoments=usemoments,
                        show_components=show_components, annotate=annotate,
                        debug=debug, use_lmfit=use_lmfit, **kwargs)
            else:
                self.peakbgfit(usemoments=usemoments, vheight=False,
                        height=0.0, annotate=annotate, use_lmfit=use_lmfit,
                        show_components=show_components, debug=debug, **kwargs)
            if self.Spectrum.plotter.autorefresh: self.Spectrum.plotter.refresh()
        else:
            if multifit:
                print "Can't fit with given fittype %s: it is not Registryed as a multifitter." % self.fittype
            else:
                print "Can't fit with given fittype %s: it is not Registryed as a singlefitter." % self.fittype
            return
        if save: self.savefit()


    def EQW(self, plot=False, plotcolor='g', annotate=False, alpha=0.5, loc='lower left'):
        """
        Returns the equivalent width (integral of "baseline" or "continuum"
        minus the spectrum) over the selected range
        """
        if np.median(self.Spectrum.baseline.basespec) == 0:
            raise ValueError("Baseline / continuum is zero: equivalent width is undefined.")
        elif np.median(self.Spectrum.baseline.basespec) < 0:
            if mycfg.WARN: warn( "WARNING: Baseline / continuum is negative: equivalent width is poorly defined." )
        diffspec = (self.Spectrum.baseline.basespec - self.Spectrum.data)
        dx = np.abs((self.Spectrum.xarr[self.xmax-1]-self.Spectrum.xarr[self.xmin]) / (self.xmax-self.xmin))
        sumofspec = diffspec[self.xmin:self.xmax].sum() * dx
        eqw = sumofspec / np.median(self.Spectrum.baseline.basespec)
        if plot:
            midpt_pixel = np.round((self.xmin+self.xmax)/2.0)
            midpt       = self.Spectrum.xarr[midpt_pixel]
            midpt_level = self.Spectrum.baseline.basespec[midpt_pixel]
            print "EQW plotting: ",midpt,midpt_pixel,midpt_level,eqw
            self.Spectrum.plotter.axis.fill_between(
                    [midpt-eqw/2.0,midpt+eqw/2.0],
                    [0,0],
                    [midpt_level,midpt_level],
                    color=plotcolor,
                    alpha=alpha,
                    label='EQW: %0.3g' % eqw)
            if annotate:
                self.Spectrum.plotter.axis.legend(
                        [(matplotlib.collections.CircleCollection([0],facecolors=[plotcolor],edgecolors=[plotcolor]))],
                        [('EQW: %0.3g' % eqw)], 
                        markerscale=0.01, borderpad=0.1, handlelength=0.1,
                        handletextpad=0.1, loc=loc)
            if self.Spectrum.plotter.autorefresh:
                self.Spectrum.plotter.refresh()
        return eqw

    def register_fitter(self,*args,**kwargs):
        """
        Register a model fitter
        """
        self.Registry.add_fitter(*args,**kwargs)

    register_fitter.__doc__ += Registry.add_fitter.__doc__
    
    def seterrspec(self,usestd=None,useresiduals=True):
        """
        Simple wrapper function to set the error spectrum; will either use the
        input spectrum or determine the error using the RMS of the residuals,
        depending on whether the residuals exist.
        """
        if (self.Spectrum.error is not None) and not usestd:
            if (self.Spectrum.error == 0).all():
                if self.residuals is not None and useresiduals: 
                    self.errspec = np.ones(self.spectofit.shape[0]) * self.residuals.std()
                elif type(self.Spectrum.error) is np.ma.masked_array:
                    # force errspec to be a non-masked array of ones
                    self.errspec = self.Spectrum.error.data + 1
                else:
                    self.errspec = self.Spectrum.error + 1
            else:
                # this is the default behavior if spectrum.error is set
                self.errspec = self.Spectrum.error.copy()
        elif self.residuals is not None and useresiduals: 
            self.errspec = np.ones(self.spectofit.shape[0]) * self.residuals.std()
        else: self.errspec = np.ones(self.spectofit.shape[0]) * self.spectofit.std()

    def setfitspec(self):
        """
        Set the spectrum that will be fit.  This is primarily to remove NANs
        from consideration: if you simply remove the data from both the X-axis
        and the Y-axis, it will not be considered for the fit, and a linear
        X-axis is not needed for fitting.

        However, it may be possible to do this using masked arrays instead of
        setting errors to be 1e10....
        """
        if self.Spectrum.data.sum() is np.ma.masked:
            self.spectofit = np.zeros_like(self.Spectrum.data)
            self.errspec = np.zeros_like(self.Spectrum.data)
            self._valid = False
            return
        self.spectofit = np.copy(self.Spectrum.data)
        self._valid = True
        if hasattr(self.Spectrum,'baseline'):
            if (self.Spectrum.baseline.subtracted is False 
                    and self.Spectrum.baseline.basespec is not None
                    and len(self.spectofit) == len(self.Spectrum.baseline.basespec)):
                self.spectofit -= self.Spectrum.baseline.basespec
        OKmask = (self.spectofit==self.spectofit)
        self.spectofit[(True-OKmask)] = 0
        self.seterrspec()
        self.errspec[(True-OKmask)] = 1e10
        if self.includemask is not None and (self.includemask.shape == self.errspec.shape):
            self.errspec[True - self.includemask] = 1e10

    def multifit(self, fittype=None, renormalize='auto', annotate=None,
            show_components=None, verbose=True, color=None, 
            use_window_limits=None, use_lmfit=False, **kwargs):
        """
        Fit multiple gaussians (or other profiles)

        fittype - What function will be fit?  fittype must have been Registryed in the
            singlefitters dict.  Uses default ('gaussian') if not specified
        renormalize - if 'auto' or True, will attempt to rescale small data (<1e-9) to be 
            closer to 1 (scales by the median) so that the fit converges better
        """
        self.setfitspec()
        if not self._valid:
            raise ValueError("Data are invalid; cannot be fit.")
        #if self.fitkwargs.has_key('negamp'): self.fitkwargs.pop('negamp') # We now do this in gaussfitter.py
        if fittype is not None: self.fittype = fittype
        if len(self.guesses) < self.Registry.npars[self.fittype]:
            raise ValueError("Too few parameters input.  Need at least %i for %s models" % (self.Registry.npars[self.fittype],self.fittype))
        self.npeaks = len(self.guesses)/self.Registry.npars[self.fittype]
        self.fitter = self.Registry.multifitters[self.fittype]
        self.vheight = False

        # add kwargs to fitkwargs
        self.fitkwargs.update(kwargs)
        if 'renormalize' in self.fitkwargs:
            del self.fitkwargs['renormalize']
        
        scalefactor = 1.0
        if renormalize in ('auto',True):
            datarange = self.spectofit[self.xmin:self.xmax].max() - self.spectofit[self.xmin:self.xmax].min()
            if abs(datarange) < 1e-9:
                scalefactor = np.median(np.abs(self.spectofit))
                if verbose: print "Renormalizing data by factor %e to improve fitting procedure" % scalefactor
                self.spectofit /= scalefactor
                self.errspec   /= scalefactor
                for ii in xrange(self.npeaks): # assume first parameter is amplitude
                    for jj,par in enumerate(self.fitter.parinfo):
                        if par.scaleable:
                            self.guesses[self.fitter.npars*ii+jj] /= scalefactor

        mpp,model,mpperr,chi2 = self.fitter(
                self.Spectrum.xarr[self.xmin:self.xmax], 
                self.spectofit[self.xmin:self.xmax], 
                err=self.errspec[self.xmin:self.xmax],
                npeaks=self.npeaks,
                params=self.guesses,
                use_lmfit=use_lmfit,
                **self.fitkwargs)

        self.spectofit *= scalefactor
        self.errspec   *= scalefactor

        if hasattr(self.fitter.mp,'status'):
            self.mpfit_status = models.mpfit_messages[self.fitter.mp.status]

        if model is None:
            raise ValueError("Model was not set by fitter.  Examine your fitter.")
        self.chi2 = chi2
        self.model = model * scalefactor

        self.parinfo = self.fitter.parinfo

        self.dof  = self.includemask.sum()-self.npeaks*self.Registry.npars[self.fittype]+np.sum(self.parinfo.fixed)

        # rescale any scaleable parameters
        for par in self.parinfo:
            if par.scaleable:
                par.value = par.value * scalefactor
                par.error = par.error * scalefactor

        self.modelpars = self.parinfo.values
        self.modelerrs = self.parinfo.errors
        self.residuals = self.spectofit[self.xmin:self.xmax] - self.model
        if self.Spectrum.plotter.axis is not None:
            if color is not None:
                kwargs.update({'composite_fit_color':color})
            self.plot_fit(annotate=annotate, 
                    show_components=show_components,
                    use_window_limits=use_window_limits,
                    **kwargs)
                
        # Re-organize modelerrs so that any parameters that are tied to others inherit the errors of the params they are tied to
        if 'tied' in self.fitkwargs:
            for ii, element in enumerate(self.fitkwargs['tied']):
                if not element.strip(): continue
                
                i1 = element.index('[') + 1
                i2 = element.index(']')
                loc = int(element[i1:i2])
                self.modelerrs[ii] = self.modelerrs[loc]

        # make sure the full model is populated
        self._full_model()
                
    def peakbgfit(self, usemoments=True, annotate=None, vheight=True, height=0,
            negamp=None, fittype=None, renormalize='auto', color=None, use_lmfit=False,
            show_components=None, debug=False, nsigcut_moments=None, **kwargs):
        """
        Fit a single peak (plus a background)

        usemoments - The initial guess will be set by the fitter's 'moments' function
            (this overrides 'guesses')
        annotate - Make a legend?
        vheight - Fit a (constant) background as well as a peak?
        height - initial guess for background
        negamp - If True, assumes amplitude is negative.  If False, assumes positive.  If 
            None, can be either.
        fittype - What function will be fit?  fittype must have been Registryed in the
            singlefitters dict
        renormalize - if 'auto' or True, will attempt to rescale small data (<1e-9) to be 
            closer to 1 (scales by the median) so that the fit converges better
        nsigcut_moments - pass to moment guesser; can do a sigma cut for moment guessing
        """
        self.npeaks = 1
        self.auto = True
        self.setfitspec()
        if fittype is not None: self.fittype=fittype
        if usemoments: # this can be done within gaussfit but I want to save them
            # use this INDEPENDENT of fittype for now (voigt and gauss get same guesses)
            self.guesses = self.Registry.singlefitters[self.fittype].moments(
                    self.Spectrum.xarr[self.xmin:self.xmax],
                    self.spectofit[self.xmin:self.xmax], vheight=vheight,
                    negamp=negamp, nsigcut=nsigcut_moments, **kwargs)
            #if vheight is False: self.guesses = [height]+self.guesses
        else:
            if negamp: self.guesses = [height,-1,0,1]
            else:  self.guesses = [height,1,0,1]

        NP = self.Registry.singlefitters[self.fittype].default_npars
        if NP > 3:
            for ii in xrange(3,NP):
                self.guesses += [0.0]

        self.fitter = self.Registry.singlefitters[self.fittype]

        if debug: print "n(guesses): %s  Guesses: %s  vheight: %s " % (len(self.guesses),self.guesses,vheight)

        scalefactor = 1.0
        if renormalize in ('auto',True):
            datarange = self.spectofit[self.xmin:self.xmax].max() - self.spectofit[self.xmin:self.xmax].min()
            if abs(datarange) < 1e-9:
                scalefactor = np.median(np.abs(self.spectofit))
                print "Renormalizing data by factor %e to improve fitting procedure" % scalefactor
                self.spectofit /= scalefactor
                self.errspec   /= scalefactor
                self.guesses[0] /= scalefactor
                if vheight: self.guesses[1] /= scalefactor

        if debug: print "Guesses before fit: ",self.guesses
        mpp,model,mpperr,chi2 = self.fitter(
                self.Spectrum.xarr[self.xmin:self.xmax],
                self.spectofit[self.xmin:self.xmax],
                err=self.errspec[self.xmin:self.xmax],
                vheight=vheight,
                params=self.guesses,
                debug=debug,
                use_lmfit=use_lmfit,
                **self.fitkwargs)
        if debug: print "Guesses, fits after: ",self.guesses, mpp

        self.spectofit *= scalefactor
        self.errspec   *= scalefactor
        
        if hasattr(self.fitter.mp,'status'):
            self.mpfit_status = models.mpfit_messages[self.fitter.mp.status]
        self.parinfo = self.fitter.parinfo

        if model is None:
            raise ValueError("Model was not set by fitter.  Examine your fitter.")
        self.chi2 = chi2
        self.dof  = self.includemask.sum()-self.npeaks*self.Registry.npars[self.fittype]-vheight+np.sum(self.parinfo.fixed)
        self.vheight=vheight
        if vheight: 
            self.Spectrum.baseline.baselinepars = [mpp[0]*scalefactor] # first item in list form
            self.Spectrum.baseline.basespec = self.Spectrum.data*0 + mpp[0]*scalefactor
            self.model = model*scalefactor - mpp[0]*scalefactor
            # I removed this recently for some reason, but more code depends on it being in place
            # Need to figure out *WHY* anything would want an extra parameter
            if len(mpp) == self.fitter.npars+1:
                mpp = mpp[1:]
        else: self.model = model*scalefactor
        self.residuals = self.spectofit[self.xmin:self.xmax] - self.model*scalefactor
        self.modelpars = mpp
        self.modelerrs = mpperr
        
        # rescale any scaleable parameters
        for par in self.parinfo:
            if par.scaleable:
                par.value = par.value * scalefactor
                par.error = par.error * scalefactor

        if self.Spectrum.plotter.axis is not None:
            if color is not None:
                kwargs.update({'composite_fit_color':color})
            self.plot_fit(annotate=annotate,
                show_components=show_components, **kwargs)

        # make sure the full model is populated
        self._full_model(debug=debug)

        if debug: print "Guesses, fits after vheight removal: ",self.guesses, mpp

    def _full_model(self, debug=False, **kwargs):
        """
        Compute the model for the whole spectrum
        """
        self.fullmodel = self.get_full_model(debug=debug)
        self.fullresiduals = self.Spectrum.data - self.fullmodel

    def get_full_model(self, debug=False):
        """ compute the model over the full axis """ 
        return self.get_model(self.Spectrum.xarr, debug=debug)

    def get_model(self, xarr, debug=False):
        """ Compute the model over a given axis """
        if self.Spectrum.baseline.subtracted or self.vheight:
            return self.fitter.n_modelfunc(self.parinfo,**self.fitter.modelfunc_kwargs)(xarr)
        else:
            return self.fitter.n_modelfunc(self.parinfo,**self.fitter.modelfunc_kwargs)(xarr) + self.Spectrum.baseline.get_model(xarr)

    def get_model_frompars(self, xarr, pars, debug=False):
        """ Compute the model over a given axis """
        if self.Spectrum.baseline.subtracted or self.vheight:
            return self.fitter.n_modelfunc(pars,**self.fitter.modelfunc_kwargs)(xarr)
        else:
            return self.fitter.n_modelfunc(pars,**self.fitter.modelfunc_kwargs)(xarr) + self.Spectrum.baseline.get_model(xarr)
                
    def plot_fit(self, annotate=None, show_components=None,
            composite_fit_color='red', component_fit_color='blue', lw=0.5,
            composite_lw=0.75, component_lw=0.75, component_kwargs={},
            use_window_limits=None, show_hyperfine_components=None, **kwargs):
        """
        Plot the fit.  Must have fitted something before calling this!  
        
        It will be automatically called whenever a spectrum is fit (assuming an
        axis for plotting exists)

        kwargs are passed to the fitter's components attribute
        """
        #if self.Spectrum.baseline.subtracted is False and self.Spectrum.baseline.basespec is not None:
        #    # don't display baseline if it's included in the fit
        #    plot_offset = self.Spectrum.plotter.offset+(self.Spectrum.baseline.basespec * (True-self.vheight))
        #else:
        plot_offset = self.Spectrum.plotter.offset

        self._full_model()
        self.modelplot += self.Spectrum.plotter.axis.plot(
                self.Spectrum.xarr,
                self.fullmodel + plot_offset,
                color=composite_fit_color, linewidth=lw)
        
        # Plot components
        if show_components or show_hyperfine_components:
            if show_hyperfine_components is not None:
                component_kwargs['return_hyperfine_components'] = show_hyperfine_components
            self._component_kwargs = component_kwargs
            self.modelcomponents = self.fitter.components(self.Spectrum.xarr, self.modelpars, **component_kwargs)
            for data in self.modelcomponents:
                # can have multidimensional components
                if len(data.shape) > 1:
                    for d in data:
                        self._plotted_components += self.Spectrum.plotter.axis.plot(self.Spectrum.xarr,
                            d + plot_offset,
                            color=component_fit_color, linewidth=component_lw)
                else:
                    self._plotted_components += self.Spectrum.plotter.axis.plot(self.Spectrum.xarr,
                        data + plot_offset,
                        color=component_fit_color, linewidth=component_lw)
                

        uwl = use_window_limits if use_window_limits is not None else self.use_window_limits
        plotkwargs = {}
        plotkwargs.update(self.Spectrum.plotter.plotkwargs)
        plotkwargs['use_window_limits'] = uwl
        self.Spectrum.plotter.reset_limits(**plotkwargs)
        if self.Spectrum.plotter.autorefresh:
            self.Spectrum.plotter.refresh()

        if annotate or ((annotate is None) and self.autoannotate):
            self.annotate()
            if self.vheight: self.Spectrum.baseline.annotate()

    def fullsizemodel(self):
        """
        If the gaussian was fit to a sub-region of the spectrum,
        expand it (with zeros) to fill the spectrum.  
        """

        if self.model.shape != self.Spectrum.data.shape:
            temp = np.zeros(self.Spectrum.data.shape)
            temp[self.xmin:self.xmax] = self.model
            self.model = temp
            self.residuals = self.spectofit - self.model
            self.selectregion(reset=True)

    def plotresiduals(self, fig=2, axis=None, clear=True, color='k',
            linewidth=0.5, drawstyle='steps-mid', **kwargs):
        """
        Plot residuals of the fit.  Specify a figure or
        axis; defaults to figure(2).

        kwargs are passed to matplotlib plot
        """
        self._full_model()
        if axis is None:
            if isinstance(fig,int):
                fig=matplotlib.pyplot.figure(fig)
            self.residualaxis = matplotlib.pyplot.gca()
            if clear: self.residualaxis.clear()
        else:
            self.residualaxis = axis
            if clear: self.residualaxis.clear()
        self.residualplot = self.residualaxis.plot(self.Spectrum.xarr,
                self.fullresiduals,drawstyle=drawstyle,
                linewidth=linewidth, color=color, **kwargs)
        if self.Spectrum.plotter.xmin is not None and self.Spectrum.plotter.xmax is not None:
            self.residualaxis.set_xlim(self.Spectrum.plotter.xmin,self.Spectrum.plotter.xmax)
        if self.Spectrum.plotter.ymin is not None and self.Spectrum.plotter.ymax is not None:
            self.residualaxis.set_ylim(self.Spectrum.plotter.ymin,self.Spectrum.plotter.ymax)
        self.residualaxis.set_xlabel(self.Spectrum.plotter.xlabel)
        self.residualaxis.set_ylabel(self.Spectrum.plotter.ylabel)
        self.residualaxis.set_title("Residuals")
        self.residualaxis.figure.canvas.draw()

    def annotate(self,loc='upper right',labelspacing=0.25, markerscale=0.01,
            borderpad=0.1, handlelength=0.1, handletextpad=0.1, frameon=False,
            chi2=None, **kwargs):
        """
        Add a legend to the plot showing the fitted parameters

        clearlegend() will remove the legend

        chi2 : {True or 'reduced' or 'optimal'}
        
        kwargs passed to legend
        """
        self._clearlegend()
        pl = matplotlib.collections.CircleCollection([0],edgecolors=['k'])
        if hasattr(self.fitter,'annotations'):
            self._annotation_labels = self.fitter.annotations()
        else:
            raise Exception("Fitter %s has no annotations." % self.fitter)

        #xtypename = units.unit_type_dict[self.Spectrum.xarr.xtype]
        xcharconv = units.CaseInsensitiveDict({'frequency':'\\nu', 'wavelength':'\\lambda', 'velocity':'v', 'pixels':'x'})
        xchar = xcharconv[self.Spectrum.xarr.xtype]
        self._annotation_labels = [L.replace('x',xchar) if L[1]=='x' else L for L in self._annotation_labels]

        if chi2 is not None:
            if chi2 == 'reduced':
                self._annotation_labels.append('$\\chi^2/\\nu = %0.2g$' % (self.chi2/self.dof))
            elif chi2 ==  'optimal':
                self._annotation_labels.append('$\\chi^2/\\nu = %0.2g$' % self.optimal_chi2())
            else:
                self._annotation_labels.append('$\\chi^2 = %0.2g$' % self.chi2)

        self.fitleg = self.Spectrum.plotter.axis.legend(
                tuple([pl]*len(self._annotation_labels)),
                self._annotation_labels, loc=loc, markerscale=markerscale,
                borderpad=borderpad, handlelength=handlelength,
                handletextpad=handletextpad, labelspacing=labelspacing,
                frameon=frameon, **kwargs)
        self.Spectrum.plotter.axis.add_artist(self.fitleg)
        self.fitleg.draggable(True)
        if self.Spectrum.plotter.autorefresh: self.Spectrum.plotter.refresh()

    def print_fit(self, print_baseline=True, **kwargs):
        """
        Print the best-fit parameters to the command line
        """

        if self.Spectrum.baseline.baselinepars is not None and print_baseline:
            print "Baseline: " + " + ".join(["%12g x^%i" % (x,i) for i,x in enumerate(self.Spectrum.baseline.baselinepars[::-1])])

        for i,p in enumerate(self.parinfo):
            print "%15s: %12g +/- %12g" % (p['parname'],p['value'],p['error'])

    def clear(self, legend=True, components=True):
        """
        Remove the fitted model from the plot

        Also removes the legend by default
        """
        if self.Spectrum.plotter.axis is not None:
            for p in self.modelplot:
                p.set_visible(False)
            if legend: self._clearlegend()
            if components: self._clearcomponents()
            if self.Spectrum.plotter.autorefresh: self.Spectrum.plotter.refresh()

    def _clearcomponents(self):
        for pc in self._plotted_components:
            pc.set_visible(False)
            if pc in self.Spectrum.plotter.axis.lines:
                self.Spectrum.plotter.axis.lines.remove(pc)
        if self.Spectrum.plotter.autorefresh: self.Spectrum.plotter.refresh()

    def _clearlegend(self):
        """
        Remove the legend from the plot window
        """
        if self.fitleg is not None: 
            # don't remove fitleg unless it's in the current axis self.fitleg.set_visible(False)
            if self.fitleg in self.Spectrum.plotter.axis.artists:
                self.Spectrum.plotter.axis.artists.remove(self.fitleg)
        if self.Spectrum.plotter.autorefresh: self.Spectrum.plotter.refresh()
    

    def savefit(self):
        """
        Save the fit parameters from a Gaussian fit to the FITS header
        ***THESE SHOULD BE WRITTEN FOR EACH TYPE OF MODEL TO BE FIT***
        """
        if self.modelpars is not None and hasattr(self.Spectrum,'header'):
            for ii,p in enumerate(self.modelpars):
                if ii % 3 == 0: self.Spectrum.header.update('AMP%1i' % (ii/3),p,comment="Gaussian best fit amplitude #%i" % (ii/3))
                if ii % 3 == 1: self.Spectrum.header.update('CEN%1i' % (ii/3),p,comment="Gaussian best fit center #%i" % (ii/3))
                if ii % 3 == 2: self.Spectrum.header.update('WID%1i' % (ii/3),p,comment="Gaussian best fit width #%i" % (ii/3))

    def downsample(self,factor):
        """
        Downsample the model spectrum (and the spectofit spectra)
        This should only be done when Spectrum.smooth is called
        """
        if self.model is not None:
            self.model = self.model[::factor]
            self.residuals = self.residuals[::factor]
        self.spectofit = self.spectofit[::factor]
        self.errspec = self.errspec[::factor]

    def crop(self,x1pix,x2pix):
        """
        When spectrum.crop is called, this must be too
        """
        if self.model is not None:
            self.model = self.model[x1pix:x2pix]
        if hasattr(self,'fullmodel'):
            self.fullmodel = self.fullmodel[x1pix:x2pix]
        self.includemask = self.includemask[x1pix:x2pix]

    def integral(self, direct=False, threshold='auto', integration_limits=[],
            return_error=False, **kwargs):
        """
        Return the integral of the fitted spectrum

        if direct=True, return the integral of the spectrum over a range
        defined by the threshold or integration limits if defined

        note that integration_limits will operate directly on the DATA, which means that
        if you've baselined without subtract=True, the baseline will be included in the integral

        if return_error is set, the error computed by
        sigma = sqrt(sum(sigma_i^2)) * dx
        will be returned as well
        """

        dx = self.Spectrum.xarr.cdelt()
        if dx is None:
            dx = np.abs(np.concatenate([np.diff(self.Spectrum.xarr),[0]]))
            warn("Irregular X-axis.  The last pixel is ignored.")
        else:
            # shouldn't shape be a 'propery'
            dx = np.repeat(np.abs(dx), self.Spectrum.shape)

        if direct:
            self._full_model()
            if len(integration_limits) == 2:
                x1 = np.argmin(np.abs(integration_limits[0]-self.Spectrum.xarr))
                x2 = np.argmin(np.abs(integration_limits[1]-self.Spectrum.xarr))
                if x1>x2: x1,x2 = x2,x1
                integ = (self.Spectrum.data[x1:x2] * dx[x1:x2]).sum()
                if return_error:
                    # compute error assuming a "known mean" (not a sample mean).  If sample mean, multiply
                    # by sqrt(len(dx)/(len(dx)-1))  (which should be very near 1)
                    error = np.sqrt((dx[x1:x2] * self.Spectrum.error[x1:x2]**2).sum() / dx[x1:x2].sum())
                    return integ,error
                else:
                    return integ
            elif threshold=='auto':
                threshold = 0.01 * np.abs( self.fullmodel ).max()

            OK = np.abs( self.fullmodel ) > threshold
            integ = (self.spectofit[OK] * dx[OK]).sum()
            error = np.sqrt((self.errspec[OK]**2 * dx[OK]).sum()/dx[OK].sum())
        else:
            if not hasattr(self.fitter,'integral'):
                raise AttributeError("The fitter %s does not have an integral implemented" % self.fittype)

            if self.Spectrum.xarr.cdelt() is not None:
                dx = np.median(dx)
                integ = self.fitter.integral(self.modelpars, **kwargs) * dx
                if return_error:
                    if mycfg.WARN: print "WARNING: The computation of the error on the integral is not obviously correct or robust... it's just a guess."
                    self._full_model()
                    OK = np.abs( self.fullmodel ) > threshold
                    error = np.sqrt((self.errspec[OK]**2).sum()) * dx
                    #raise NotImplementedError("We haven't written up correct error estimation for integrals of fits")
            else:
                integ = 0
                error = 0
                warn("An analytic integal could not be computed.  Try direct=True when integrating, or find a way to linearize the X-axis")
        if return_error:
            return integ,error
        else:
            return integ

    def shift_pars(self, frame=None):
        """
        Shift the velocity / wavelength / frequency of the fitted parameters
        into a different frame

        Right now this only takes care of redshift and only if redshift is defined.
        It should be extended to do other things later
        """
        for ii,pi in enumerate(self.parinfo):
            for partype in ('shift','offset','velo'):
                if partype in str.lower(pi['parname']):
                    if frame is not None:
                        self.modelpars[ii] = self.Spectrum.xarr.x_in_frame(self.modelpars[ii], frame)

    def moments(self, **kwargs):
        """
        Return the moments 

        see :mod:`moments`
        """
        return self.Registry.singlefitters[self.fittype].moments(
                self.Spectrum.xarr[self.xmin:self.xmax],
                self.spectofit[self.xmin:self.xmax],  **kwargs)

    def button3action(self, event, debug=False):
        """
        Disconnect the interactiveness
        Perform the fit (or die trying)
        Hide the guesses
        """
        self.Spectrum.plotter.figure.canvas.mpl_disconnect(self.click)
        self.Spectrum.plotter.figure.canvas.mpl_disconnect(self.keyclick)
        if self.npeaks > 0:
            print len(self.guesses)/3," Guesses: ",self.guesses," X channel range: ",self.xmin,self.xmax
            if len(self.guesses) % 3 == 0:
                self.multifit(use_window_limits=True)
                for p in self.button2plot + self.button1plot:
                    p.set_visible(False)
            else: 
                print "error, wrong # of pars"

        # disconnect interactive window (and more importantly, reconnect to
        # original interactive cmds)
        self.clear_all_connections()

    def copy(self, parent=None):
        """
        Create a copy of the spectral fit - includes copies of the _full_model,
        the registry, the fitter, parinfo, modelpars, modelerrs, model, npeaks

        [ parent ] 
            A spectroscopic axis instance that is the parent of the specfit
            instance.  This needs to be specified at some point, but defaults
            to None to prevent overwriting a previous plot.
        """

        newspecfit = copy.copy(self)
        newspecfit.Spectrum = parent
        newspecfit.parinfo = copy.deepcopy(self.parinfo)
        if newspecfit.parinfo is None:
            newspecfit.modelpars = None
            newspecfit.modelerrs = None
        else:
            newspecfit.modelpars = newspecfit.parinfo.values
            newspecfit.modelerrs = newspecfit.parinfo.errors
        newspecfit.includemask = self.includemask.copy() 
        newspecfit.model = copy.copy( self.model )
        newspecfit.npeaks = self.npeaks
        if hasattr(self,'fitter'):
            newspecfit.fitter = copy.copy( self.fitter )
            newspecfit.fitter.parinfo = newspecfit.parinfo
        if hasattr(self,'fullmodel'):
            newspecfit._full_model()

        # this is ridiculous, absurd, and infuriating...
        newspecfit.button2action = newspecfit.guesspeakwidth

        if parent is not None:
            newspecfit.Spectrum.plotter = parent.plotter
        else:
            newspecfit.Spectrum.plotter = None

        return newspecfit

    def add_sliders(self, parlimitdict=None, **kwargs):
        """
        Add a Sliders window in a new figure appropriately titled

        *parlimitdict* [ dict ]
            Each parameter needs to have displayed limits; these are set in
            min-max pairs.  If this is left empty, the widget will try to guess
            at reasonable limits, but the guessing is not very sophisticated
            yet.

        .. TODO:: Add a button in the navbar that makes this window pop up
        http://stackoverflow.com/questions/4740988/add-new-navigate-modes-in-matplotlib
        """
        import widgets

        if parlimitdict is None:
            # try to create a reasonable parlimit dict
            parlimitdict = {}

        for param in self.parinfo:
            if not param.parname in parlimitdict:
                if any( (x in param['parname'].lower() for x in ('shift','xoff')) ):
                    lower,upper = (self.Spectrum.xarr[self.includemask].min(),self.Spectrum.xarr[self.includemask].max())
                elif any( (x in param['parname'].lower() for x in ('width','fwhm')) ):
                    xvalrange = (self.Spectrum.xarr[self.includemask].max()-self.Spectrum.xarr[self.includemask].min())
                    lower,upper = (0,xvalrange)
                elif any( (x in param['parname'].lower() for x in ('amp','peak','height')) ):
                    datarange = self.spectofit.max() - self.spectofit.min()
                    lower,upper = (param['value']-datarange, param['value']+datarange)
                else:
                    lower = param['value'] * 0.1
                    upper = param['value'] * 10

                # override guesses with limits
                if param.limited[0]:
                    lower = param.limits[0]
                if param.limited[1]:
                    upper = param.limits[1]

                parlimitdict[param.parname] = (lower,upper)

        if hasattr(self,'fitter'):
            self.SliderWidget = widgets.FitterSliders(self, self.Spectrum.plotter.figure, npars=self.fitter.npars, parlimitdict=parlimitdict, **kwargs)
        else:
            print "Must have a fitter instantiated before creating sliders"

    def optimal_chi2(self, reduced=True):
        """
        Compute an "optimal" chi^2 statistic, i.e. one in which only pixels in
        which the model is statistically significant are included

        Parameters
        ----------
        reduced : bool
            Return the reduced chi^2

        Returns
        -------
        Chi^2 statistic or reduced chi^2 statistic
        `\\chi^2 = \\sum( (d_i - m_i)^2 / e_i^2 )
        """

        modelmask = np.abs(self.fullmodel) > self.errspec

        chi2 = np.sum( (self.fullresiduals[modelmask]/self.errspec[modelmask])**2 )

        if reduced:
            dof = modelmask.sum() - self.fitter.npars - self.vheight + np.sum(self.parinfo.fixed)  # vheight included here or not?
            return chi2/dof
        else:
            return chi2

    def get_pymc(self, **kwargs):
        """
        Create a pymc MCMC sampler from the current fitter.  Defaults to 'uninformative' priors

        `kwargs` are passed to the fitter's get_pymc method, with parameters defined below.

        Parameters
        ----------
        data : np.ndarray
        error : np.ndarray
        use_fitted_values : bool
            Each parameter with a measured error will have a prior defined by
            the Normal distribution with sigma = par.error and mean = par.value

        Examples
        --------

        >>> x = pyspeckit.units.SpectroscopicAxis(np.linspace(-10,10,50), unit='km/s')
        >>> e = np.random.randn(50)
        >>> d = np.exp(-np.asarray(x)**2/2.)*5 + e
        >>> sp = pyspeckit.Spectrum(data=d, xarr=x, error=np.ones(50)*e.std())
        >>> sp.specfit(fittype='gaussian')
        >>> MCuninformed = sp.specfit.get_pymc()
        >>> MCwithpriors = sp.specfit.get_pymc(use_fitted_values=True)
        >>> MCuninformed.sample(1000)
        >>> MCuninformed.stats()['AMPLITUDE0']
        >>> # WARNING: This will fail because width cannot be set <0, but it may randomly reach that...
        >>> # How do you define a likelihood distribution with a lower limit?!
        >>> MCwithpriors.sample(1000)
        >>> MCwithpriors.stats()['AMPLITUDE0']
        """
        if hasattr(self.fitter,'get_pymc'):
            return self.fitter.get_pymc(self.Spectrum.xarr, self.spectofit, self.errspec, **kwargs)

    def get_emcee(self, nwalkers=None, **kwargs):
        """
        Get an emcee walker ensemble for the data & model using the current model type

        Parameters
        ----------
        data : np.ndarray
        error : np.ndarray
        nwalkers : int
            Number of walkers to use.  Defaults to 2 * self.fitters.npars

        Examples
        --------

        >>> import pyspeckit
        >>> x = pyspeckit.units.SpectroscopicAxis(np.linspace(-10,10,50), unit='km/s')
        >>> e = np.random.randn(50)
        >>> d = np.exp(-np.asarray(x)**2/2.)*5 + e
        >>> sp = pyspeckit.Spectrum(data=d, xarr=x, error=np.ones(50)*e.std())
        >>> sp.specfit(fittype='gaussian')
        >>> emcee_ensemble = sp.specfit.get_emcee()
        >>> p0 = emcee_ensemble.p0 * (np.random.randn(*emcee_ensemble.p0.shape) / 10. + 1.0)
        >>> pos,logprob,state = emcee_ensemble.run_mcmc(p0,100)
        """
        if hasattr(self.fitter,'get_emcee_ensemblesampler'):
            nwalkers = (self.fitter.npars * self.fitter.npeaks + self.fitter.vheight) * 2
            emc = self.fitter.get_emcee_ensemblesampler(self.Spectrum.xarr, self.spectofit, self.errspec, nwalkers)
            emc.nwalkers = nwalkers
            emc.p0 = np.array([self.parinfo.values] * emc.nwalkers)
            return emc



"""
GBTIDL SDFITS file
==================

GBTIDL SDFITS files representing GBT observing sessions can be read into
pyspeckit.  Additional documentation is needed.  Nodding reduction is
supported, frequency switching is not.

"""
import pyfits
import pyspeckit
import numpy as np
try:
    import coords
except ImportError:
    coords = None

def unique_targets(sdfitsfile):
    bintable = _get_bintable(sdfitsfile)
    return uniq(bintable.data['OBJECT'])

def count_integrations(sdfitsfile, target):
    """
    Return the number of integrations for a given target
    (uses one sampler; assumes same number for all samplers)
    """
    bintable = _get_bintable(sdfitsfile)

    whobject = bintable.data['OBJECT'] == target
    any_sampler = bintable.data['SAMPLER'][whobject][0]
    whsampler = bintable.data['SAMPLER'][whobject] == any_sampler

    return (whsampler).sum()


def list_targets(sdfitsfile, doprint=True):
    """
    List the targets, their location on the sky...
    """
    bintable = _get_bintable(sdfitsfile)

    # Things to include:
    #   Scan           Source      Vel    Proc Seqn   RestF nIF nInt nFd     Az    El
    #        1      G33.13-0.09     87.4     Nod    1  14.488   4   15   2  209.7  47.9
    strings = [ "\n%18s  %10s %10s %26s%8s %9s %9s %9s" % ("Object Name","RA","DEC","%12s%14s"%("RA","DEC"),"N(ptgs)","Exp.Time","requested","n(ints)") ]
    for objectname in uniq(bintable.data['OBJECT']):
        whobject = bintable.data['OBJECT'] == objectname
        RA,DEC = bintable.data['TRGTLONG'][whobject],bintable.data['TRGTLAT'][whobject]
        RADEC = zip(RA,DEC)
        midRA,midDEC = np.median(RA),np.median(DEC)
        npointings = len(set(RADEC))
        if coords is not None:
            sexagesimal = coords.Position((midRA,midDEC)).hmsdms()
        else:
            sexagesimal = ""
        firstsampler = bintable.data['SAMPLER'][whobject][0]
        whfirstsampler = bintable.data['SAMPLER'] == firstsampler
        exptime = bintable.data['EXPOSURE'][whobject*whfirstsampler].sum()
        duration = bintable.data['DURATION'][whobject*whfirstsampler].sum()
        n_ints = count_integrations(bintable, objectname)
        strings.append( "%18s  %10f %10f %26s%8i %9g %9g %9i" % (objectname,midRA,midDEC,sexagesimal, npointings, exptime, duration, n_ints) )

    if doprint:
        print "\n".join(strings)

    return strings

def read_gbt_scan(sdfitsfile, obsnumber=0):
    """
    Read a single scan from a GBTIDL SDFITS file
    """

    bintable = _get_bintable(sdfitsfile)

    data = bintable.data[obsnumber]['DATA']

    header = pyfits.Header()

    for par in bintable.data.dtype.names:
        if par not in ('DATA',):
            try:
                header.update(par[:8], bintable.data[obsnumber][par])
            except ValueError:
                header.update(par[:8], str(bintable.data[obsnumber][par]))
    header.update('CUNIT1','Hz')

    HDU = pyfits.PrimaryHDU(data=data,header=header)

    sp = pyspeckit.Spectrum(HDU,filetype='pyfits')
    #sp.xarr.frame = 'topo' # sp.header.get('VELDEF') #'topo'

    # HACK - temporary!
    # Convert xarr to LSR units
    #sp.xarr.convert_to_unit('m/s')
    obsfreq = header['OBSFREQ']
    centerfreq = pyspeckit.spectrum.units.SpectroscopicAxis(np.float(obsfreq),'Hz',refX=obsfreq,refX_units='Hz')
    delta_freq = (centerfreq.as_unit('m/s') + header['VFRAME']).as_unit(centerfreq.units) - centerfreq
    sp.xarr -= delta_freq

    return sp

def read_gbt_target(sdfitsfile, objectname, verbose=False):
    """
    Give an object name, get all observations of that object as an 'obsblock'
    """

    bintable = _get_bintable(sdfitsfile)

    whobject = bintable.data['OBJECT'] == objectname
    if verbose:
        print "Number of individual scans for Object %s: %i" % (objectname,whobject.sum())

    calON = bintable.data['CAL'] == 'T'
    # HACK: apparently bintable.data can sometimes treat itself as scalar...
    if np.isscalar(calON):
        calON = np.array([(val in ['T',True]) for val in bintable.data['CAL']])

    n_nods = np.unique(bintable.data['PROCSIZE'])

    blocks = {}
    for sampler in np.unique(bintable.data[whobject]['SAMPLER']):
        whsampler = bintable.data['SAMPLER'] == sampler
        nods = np.unique(bintable.data['PROCSEQN'][whsampler*whobject])
        for nod in nods:
            whnod = bintable.data['PROCSEQN'] == nod
            for onoff in ('ON','OFF'):
                calOK = (calON - (onoff=='OFF'))
                whOK = (whobject*whsampler*calOK*whnod)
                if whOK.sum() == 0:
                    continue
                if verbose:
                    print "Number of spectra for sampler %s, nod %i, cal%s: %i" % (sampler,nod,onoff,whOK.sum())
                crvals = bintable.data[whOK]['CRVAL1']
                if len(crvals) > 1:
                    maxdiff = np.diff(crvals).max()
                else:
                    maxdiff = 0
                freqres = np.max(bintable.data[whOK]['FREQRES'])
                if maxdiff < freqres:
                    splist = [read_gbt_scan(bintable,ii) for ii in np.where(whOK)[0]]
                    blocks[sampler+onoff+str(nod)] = pyspeckit.ObsBlock(splist,force=True)
                    blocks[sampler+onoff+str(nod)]._arithmetic_threshold = np.diff(blocks[sampler+onoff+str(nod)].xarr).min() / 5.
                else:
                    print "Maximum frequency difference > frequency resolution: %f > %f" % (maxdiff, freqres)

    return blocks

def reduce_gbt_target(sdfitsfile, objectname, verbose=False):
    """
    Wrapper - read an SDFITS file, get an object, reduce it (assuming nodded) and return it
    """
    # for efficiency, this should be stored - how?
    blocks = read_gbt_target(sdfitsfile, objectname, verbose=verbose)

    reduced_nods = reduce_blocks(blocks)

    return reduced_nods

def reduce_blocks(blocks, verbose=False, average=True):
    """
    Do a nodded on/off observation given a dict of observation blocks as
    produced by read_gbt_target
    """

    # strip off trailing digit to define pairs
    nodpairs = uniq([s[:-1].replace("ON","").replace("OFF","") for s in blocks])

    reduced_nods = {}

    # for each pair...
    for sampname in nodpairs:
        on1 = sampname+"ON1"
        off1 = sampname+"OFF1"
        on2 = sampname+"ON2"
        off2 = sampname+"OFF2"

        feednumber = blocks[on1].header.get('FEED')
        # don't need this - if feednumber is 1, ref feed should be 2, and vice-versa
        reference_feednumber = blocks[on1].header.get('SRFEED')
        
        on1avg = blocks[on1].average()
        off1avg = blocks[off1].average()
        on2avg = blocks[on2].average()
        off2avg = blocks[off2].average()

        # first find TSYS
        tsys1 = dcmeantsys(on1avg,off1avg,on1avg.header.get('TCAL'))
        tsys2 = dcmeantsys(on2avg,off2avg,on2avg.header.get('TCAL'))
        if verbose:
            print "Nod Pair %s (feed %i) has tsys1=%f tsys2=%f" % (sampname, feednumber, tsys1, tsys2)

        # then get the total power
        if average:
            tp1 = totalpower(on1avg, off1avg, average=False)
            tp2 = totalpower(on2avg, off2avg, average=False)
        else:
            tp1 = totalpower(blocks[on1], blocks[off1], average=False)
            tp2 = totalpower(blocks[on2], blocks[off2], average=False)

        # then do the signal-reference bit
        if feednumber == 1:
            nod = sigref(tp1,tp2,tsys2)
        elif feednumber == 2:
            nod = sigref(tp2,tp1,tsys1)

        reduced_nods[sampname] = nod

    return reduced_nods

def _get_bintable(sdfitsfile):
    """
    Private function: given a filename, HDUlist, or bintable, return a bintable
    """

    if isinstance(sdfitsfile, pyfits.HDUList):
        bintable = sdfitsfile[1]
    elif isinstance(sdfitsfile, pyfits.BinTableHDU):
        bintable = sdfitsfile
    elif isinstance (sdfitsfile, pyfits.FITS_rec):
        bintable = sdfitsfile
    else:
        bintable = pyfits.open(sdfitsfile)[1]

    return bintable

def dcmeantsys(calon,caloff,tcal,debug=False):
    """
    from GBTIDL's dcmeantsys.py
    ;  mean_tsys = tcal * mean(nocal) / (mean(withcal-nocal)) + tcal/2.0
    """

    # Use the inner 80% of data to calculate mean Tsys
    nchans = calon.data.shape[0]
    pct10 = nchans/10
    pct90 = nchans - pct10

    meanoff = np.mean(caloff.slice(pct10,pct90,units='pixels').data)
    meandiff = np.mean(calon.slice(pct10,pct90,units='pixels').data - 
                        caloff.slice(pct10,pct90,units='pixels').data)

    meanTsys = ( meanoff / meandiff * tcal + tcal/2.0 )
    if debug:
        print caloff
        print caloff.slice(pct10,pct90,units='pixels')
        print calon
        print calon.slice(pct10,pct90,units='pixels')
        print "pct10: %i  pct90: %i mean1: %f mean2: %f tcal: %f tsys: %f" % (pct10,pct90,meanoff,meandiff,tcal,meanTsys)

    return meanTsys

def totalpower(calon, caloff, average=True):
    """
    Do a total-power calibration of an on/off data set
    (see dototalpower.pro in GBTIDL)
    """

    if hasattr(calon,'average') and average:
        ON = calon.average()
    elif hasattr(calon,'average'):
        ON = calon.data
    else:
        ON = calon

    if hasattr(caloff,'average') and average:
        OFF = caloff.average()
    elif hasattr(caloff,'average'):
        OFF = caloff.data
    else:
        OFF = caloff

    TP = (ON+OFF)/2.

    return TP

def sigref(nod1, nod2, tsys_nod2):
    """
    Signal-Reference ('nod') calibration
    ; ((dcsig-dcref)/dcref) * dcref.tsys 
    see GBTIDL's dosigref
    """

    return (nod1-nod2)/nod2*tsys_nod2

def find_matched_freqs(reduced_blocks, debug=False):
    """
    Use frequency-matching to find which samplers observed the same parts of the spectrum

    *WARNING* These IF numbers don't match GBTIDL's!  I don't know how to get those to match up!
    """

    # IF order 
    sampler_numbers = [int(name[1:]) for name in reduced_blocks.keys()]
    sorted_pairs = sorted(zip(sampler_numbers,reduced_blocks.keys()))
    sorted_names = zip(*sorted_pairs)[1]

    # how many IFs?
    frequencies = dict((name,reduced_blocks[name].header.get('OBSFREQ')) for name in sorted_names)
    if debug: print frequencies
    round_frequencies = dict((name,
        round_to_resolution(reduced_blocks[name].header.get('OBSFREQ'),
                            reduced_blocks[name].header.get('FREQRES')))
                 for name in sorted_names)
    if debug: print round_frequencies
    unique_frequencies = uniq(round_frequencies.values()) # uniq is an order-preserving function
    if debug: print unique_frequencies
    nIFs = len(unique_frequencies)
    if debug:
        print frequencies.keys()
        print round_frequencies.keys()
        print unique_frequencies

    IF_dict = {}
    # most annoying part of this whole process... (besides the obvious 1+1=23452342345 error...)
    # ifnumbers are approximately defined by the order things are written.... but only very approximately
    for ifnum,freq in enumerate(unique_frequencies):
        IF_dict[ifnum] = [sampler for (sampler,rf) in round_frequencies.iteritems() if rf == freq]
        if debug: print ifnum,freq,IF_dict[ifnum]

    return IF_dict

def uniq(seq):
    """ from http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order """
    seen = set()
    seen_add = seen.add
    return [ x for x in seq if x not in seen and not seen_add(x)]


def round_to_resolution(frequency, resolution):
    """
    kind of a hack, but round the frequency to the nearest integer multiple of the resolution,
    then multiply it back into frequency space
    """
    # doesn't really work... divisor = 10**np.floor(np.log10(resolution))
    divisor = resolution
    return (np.round(frequency/divisor)) * divisor

def find_pols(block):
    """
    Get a dictionary of the polarization for each sampler
    """
    pols = dict((name,polnum_to_pol[int(block[name].header.get('CRVAL4'))]) for name in block)
    return pols

def find_feeds(block):
    """
    Get a dictionary of the feed numbers for each sampler
    """
    feeds = dict((name,block[name].header.get('FEED')) for name in block)
    return feeds

def identify_samplers(block):
    """
    Identify each sampler with an IF number, a feed number, and a polarization
    """
    feeds = find_feeds(block)
    pols = find_pols(block)
    ifs = find_matched_freqs(block)

    IDs = dict(
            (name,
                {'pol': pols[name],
                 'feed': feeds[name],
                 'IF': [k for k,v in ifs.iteritems() if name in v][0]})
                for name in block.keys()
                )

    return IDs
    
def average_pols(block):
    """
    Average the polarizations for each feed in each IF
    """

    # will have names like "(IFnum)(feed)"
    averaged_pol_dict = {}

    ifdict = find_matched_freqs(block)
    feeddict = find_feeds(block)
    IDs = identify_samplers(block)

    for ifnum,ifsampler in ifdict.iteritems():
        for sampler,feednum in feeddict.iteritems():
            if sampler not in ifsampler:
                continue
            newname = "if%ifd%i" % (ifnum, feednum)
            matched_samplers = [sampler_name for (sampler_name,ID) in IDs.iteritems()
                    if ID['feed'] == feednum and ID['IF'] == ifnum]
            if len(matched_samplers) != 2:
                raise ValueError("Too few/many matches: %s" % matched_samplers)

            if newname not in averaged_pol_dict:
                average = (block[matched_samplers[0]] + block[matched_samplers[1]]) / 2.
                averaged_pol_dict[newname] = average

    return averaged_pol_dict

def average_IF(block, debug=False):
    """
    Average the polarizations for each feed in each IF
    """

    # will have names like "(IFnum)"
    averaged_dict = {}

    ifdict = find_matched_freqs(block, debug=debug)

    import operator
    for ifnum,ifsamplers in ifdict.iteritems():
        if debug: print "if%i: freq %g" % (ifnum, block[ifsamplers[0]].header['OBSFREQ'])
        averaged_dict["if%i" % ifnum] = reduce(operator.add,[block[name] for name in ifsamplers]) / len(ifsamplers)

    return averaged_dict


polnum_to_pol = {
    1: 'I',
    2: 'Q',
    3: 'U',
    4: 'V',
   -1: 'RR',
   -2: 'LL',
   -3: 'RL',
   -4: 'LR',
   -5: 'XX',
   -6: 'YY',
   -7: 'XY',
   -8: 'YX'}


class GBTSession(object):
    """
    A class wrapping all of the above features
    """
    def __init__(self, sdfitsfile):
        """
        Load an SDFITS file or a pre-loaded FITS file
        """
        self.bintable = _get_bintable(sdfitsfile)
        self.targets = dict((target,None) for target in unique_targets(self.bintable))
        self.print_header = "\n".join( ("Observer: " + self.bintable.data[0]['OBSERVER'],
            "Project: %s" % self.bintable.header.get('PROJID'),
            "Backend: %s" % self.bintable.header.get('BACKEND'),
            "Telescope: %s" % self.bintable.header.get('TELESCOP'),
            "Bandwidth: %s" % self.bintable.data[0]['BANDWID'],
            "Date: %s" % self.bintable.data[0]['DATE-OBS']) )

    def __repr__(self):
        self.instance_info = super(GBTSession,self).__repr__()
        if not hasattr(self,'StringDescription'):
            self.StringDescription = list_targets(self.bintable, doprint=False)
        return self.print_header+"\n"+"\n".join(self.StringDescription)
    
    def __str__(self):
        if not hasattr(self,'StringDescription'):
            self.StringDescription = list_targets(self.bintable, doprint=False)
        return self.print_header+"\n"+"\n".join(self.StringDescription)

    def __getitem__(self, ind):
        return self.targets[ind]

    def load_target(self, target, **kwargs):
        """
        Load a Target...
        """
        self.targets[target] = GBTTarget(self, target)
        return self.targets[target]

    def reduce_target(self, target, **kwargs):
        """
        Reduce the data for a given object name
        """
        if self.targets[target] is None:
            self.load_target(target)

        self.targets[target].reduce(**kwargs)

        return self.targets[target]

    def reduce_all(self):
        """

        """
        for target in self.targets:
            self.reduce_target(target)

class GBTTarget(object):
    """
    A collection of ObsBlocks or Spectra
    """
    def __init__(self, Session, target, **kwargs):
        """
        Container for the individual scans of a target from a GBT session
        """
        self.name = target
        self.Session = Session
        self.blocks = read_gbt_target(Session.bintable, target, **kwargs)
        self.spectra = {}

    def __getitem__(self, ind):
        return self.spectra[ind]

    def __repr__(self):
        self.instance_info = super(GBTTarget,self).__repr__()
        self.StringDescription = [("Object %s with %i scan blocks and %i 'reduced' spectra" %
                (self.name,len(self.blocks),len(self.spectra)))]
        if hasattr(self,'reduced_scans'):
            self.StringDescription += ["%s" % ID for ID in identify_samplers(self.reduced_scans)]
        return "\n".join(self.StringDescription)

    def reduce(self, obstype='nod', **kwargs):
        """
        Reduce nodded observations (they should have been read in __init__)
        """
        self.reduced_scans = reduce_blocks(self.blocks, **kwargs)
        self.spectra.update(self.reduced_scans)

    def average_pols(self):
        if hasattr(self,'reduced_scans'):
            self.averaged_pols = average_pols(self.reduced_scans)
            self.spectra.update(self.averaged_pols)
        else:
            raise ValueError("Must reduce the scans before averaging pols")

    def average_IFs(self, **kwargs):
        if hasattr(self,'reduced_scans'):
            self.averaged_IFs = average_IF(self.reduced_scans, **kwargs)
            self.spectra.update(self.averaged_IFs)
        else:
            raise ValueError("Must reduce the scans before averaging IFs")

"""
TEST CODE
for name in reduced_nods:
    for num in '1','2':

name='A10'
num='1'
tsys = 10
while (tsys>5):
    av1 = blocks[name+'ON'+num].average()
    av2 = blocks[name+'OFF'+num].average()
    tsys = dcmeantsys(av1, av2, blocks[name+'OFF'+num].header['TCAL'],debug=True)
    if tsys < 5:
        print "%s %s: %f" % (name,num,tsys),
        print av1,np.mean(av1.slice(409,3687,units='pixels').data)
        print av2,np.mean(av2.slice(409,3687,units='pixels').data)
        print av1-av2,np.mean(av1.slice(409,3687,units='pixels').data-av2.slice(409,3687,units='pixels').data)
        print blocks[name+'OFF'+num].header['TCAL']


"""



""" 
	Copyright (C) 2009 Sergey Koposov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import mpfit 
import re
import numpy

def mpfitexpr(func, x, y, err , start_params, check=True, full_output=False, **kw):
	"""Fit the used defined expression to the data
	Input:
	- func: string with the function definition 
	- x: x vector
	- y: y vector
	- err: vector with the errors of y
	- start_params: the starting parameters for the fit
	Output:
	- The tuple (params, yfit) with best-fit params and the values of func evaluated at x
	Keywords:
	- check: boolean parameter. If true(default) the function will be checked for sanity
	- full_output: boolean parameter. If True(default is False) then instead of best-fit parameters the mpfit object is returned
	Example:
	params,yfit=mpfitexpr('p[0]+p[2]*(x-p[1])',x,y,err,[0,10,1])
	
	If you need to use numpy functions in your function, then
		you must to use the full names of these functions, e.g.:
		numpy.sin, numpy.cos etc.
	
	This function is motivated by mpfitexpr() from wonderful MPFIT IDL package
		written by Craig Markwardt	
	
	"""

	def myfunc(p,fjac=None,x=None, y=None, err=None):
		return [0, eval('(y-(%s))/err'%func)]

	myre = "[^a-zA-Z]p\[(\d+)\]"
	r = re.compile(myre)
	maxp = -1
	for m in re.finditer(r,func):
		curp = int(m.group(1))
		maxp = curp if curp > maxp else maxp	
	if check:
		if maxp == -1: 
			raise Exception("wrong function format")
		if maxp + 1 != len(start_params):
			raise Exception("the length of the start_params != the length of the parameter verctor of the function")
	fa={'x' : x, 'y' : y,'err' : err}
	res = mpfit.mpfit(myfunc,start_params,functkw=fa,**kw)
	yfit = eval(func, globals(), {'x':x, 'p': res.params})
	if full_output:
		return (res, yfit)
	else:
		return (res.params, yfit)



"""
=================
Lorentzian Fitter
=================
"""
import numpy
from numpy.ma import median
from numpy import pi
from pyspeckit.mpfit import mpfit
from . import fitter

class LorentzianFitter(fitter.SimpleFitter):

    def __init__(self,multisingle='multi'):
        self.npars = 3
        self.npeaks = 1
        self.onepeaklorentzfit = self._fourparfitter(self.onepeaklorentzian)
        if multisingle in ('multi','single'):
            self.multisingle = multisingle
        else:
            raise Exception("multisingle must be multi or single")

    def __call__(self,*args,**kwargs):
        if self.multisingle == 'single':
            return self.onepeaklorentzfit(*args,**kwargs)
        elif self.multisingle == 'multi':
            return self.multilorentzfit(*args,**kwargs)


    def onedlorentzian(x,H,A,dx,w):
        """
        Returns a 1-dimensional gaussian of form
        H+A*numpy.exp(-(x-dx)**2/(2*w**2))
        """
        return H+A/(2*pi)*w/((x-dx)**2 + (w/2.0)**2)

    def n_lorentzian(pars=None,a=None,dx=None,width=None):
        """
        Returns a function that sums over N lorentzians, where N is the length of
        a,dx,sigma *OR* N = len(pars) / 3

        The background "height" is assumed to be zero (you must "baseline" your
        spectrum before fitting)

        pars  - a list with len(pars) = 3n, assuming a,dx,sigma repeated
        dx    - offset (velocity center) values
        width - line widths (Lorentzian FWHM)
        a     - amplitudes
        """
        if len(pars) % 3 == 0:
            a = [pars[ii] for ii in xrange(0,len(pars),3)]
            dx = [pars[ii] for ii in xrange(1,len(pars),3)]
            width = [pars[ii] for ii in xrange(2,len(pars),3)]
        elif not(len(dx) == len(width) == len(a)):
            raise ValueError("Wrong array lengths! dx: %i  width %i  a: %i" % (len(dx),len(width),len(a)))

        def L(x):
            v = numpy.zeros(len(x))
            for i in range(len(dx)):
                v += a[i] / (2*pi) * w / ((x-dx)**2 + (w/2.0)**2)
            return v
        return L

    def multilorentzfit(self):
        """
        not implemented
        """
        print "Not implemented"



"""
====================
N2H+ fitter wrapper
====================

Wrapper to fit N2H+ using RADEX models.  This is meant to be used from the command line, e.g.:

    python n2hp_wrapper.py file.fits

and therefore has no independently defined functions.
"""
import pyspeckit
import numpy as np
import pyfits
from pyspeckit.spectrum import models
from pyspeckit.spectrum.models import n2hp

# EDIT THIS LINE:
path_to_radex = '/Users/adam/work/n2hp/'

# create the n2hp Radex fitter
# This step cannot be easily generalized: the user needs to read in their own grids
texgrid1 = pyfits.getdata(path_to_radex+'1-2_T=5to55_lvg_tex1.fits')
taugrid1 = pyfits.getdata(path_to_radex+'1-2_T=5to55_lvg_tau1.fits')
texgrid2 = pyfits.getdata(path_to_radex+'1-2_T=5to55_lvg_tex2.fits')
taugrid2 = pyfits.getdata(path_to_radex+'1-2_T=5to55_lvg_tau2.fits')
hdr    = pyfits.getheader(path_to_radex+'1-2_T=5to55_lvg_tau2.fits')

# this deserves a lot of explanation:
# models.n2hp.n2hp_radex is the MODEL that we are going to fit
# models.model.SpectralModel is a wrapper to deal with parinfo, multiple peaks,
# and annotations
# all of the parameters after the first are passed to the model function 
n2hp_radex_fitter = models.model.SpectralModel(
        models.n2hp.n2hp_radex, 4,
        parnames=['density','column','center','width'], 
        parvalues=[4,12,0,1],
        parlimited=[(True,True), (True,True), (False,False), (True,False)], 
        parlimits=[(1,8), (11,16), (0,0), (0,0)],
        parsteps=[0.01,0.01,0,0],
        fitunits='Hz',
        texgrid=((4,5,texgrid1),(14,15,texgrid2)), # specify the frequency range over which the grid is valid (in GHz)
        taugrid=((4,5,taugrid1),(14,15,taugrid2)),
        hdr=hdr,
        shortvarnames=("n","N","v","\\sigma"), # specify the parameter names (TeX is OK)
        grid_vwidth_scale=False,
        )


if __name__ == "__main__":

    import optparse
    parser=optparse.OptionParser()
    parser.add_option("--scalekeyword",help="Scale the data before fitting?",default='ETAMB')
    parser.add_option("--vmin",help="Mininum velocity to include",default=-50)
    parser.add_option("--vmax",help="Maximum velocity to include",default=50)
    parser.add_option("--vguess",help="Velocity guess",default=3.75)
    parser.add_option("--smooth",help="Smooth the spectrum",default=None)

    options,args = parser.parse_args()

    if len(args) == 1:
        fn = args[0]
    else:
        print "Couldn't load a file!  Crash time!"


    sp = pyspeckit.Spectrum(fn,scale_keyword=options.scalekeyword)

    sp.xarr.convert_to_unit('km/s')
    sp.crop(options.vmin,options.vmax)
    if options.smooth: 
        sp.smooth(options.smooth) 

    sp.specfit() # determine errors
    sp.error = np.ones(sp.data.shape)*sp.specfit.residuals.std()
    sp.baseline(excludefit=True)

    sp.Registry.add_fitter('n2hp_radex',
            n2hp_radex_fitter,4,multisingle='multi')
    sp.Registry.add_fitter('n2hp_vtau',
            n2hp.n2hp_vtau_fitter,4,multisingle='multi')

    sp.plotter(figure=1)
    sp.specfit(fittype='n2hp_radex',multifit=True,guesses=[4,12,options.vguess,0.43],quiet=False)

    sp.plotter(figure=2)
    sp.specfit(fittype='n2hp_vtau',multifit=True,guesses=[15,2,options.vguess,0.43],quiet=False)



import pyspeckit
import pyfits
import numpy as np

def read_galex(fitsfilename, orderselection='obj'):
    """
    Read in a GALEX xg-gsp.fits file 
    Definition of the file is here: 
    `<http://www.galex.caltech.edu/DATA/gr1_docs/GR1_Column_descriptions_for_-xg-gsp.fits.htm>`_

    Parameters
    ----------
    fitsfilename: string 
        The file name (must be a valid FITS file)
    orderselection: {'obj','objs'}
        'obj' selects the NUV 1st and FUV 2nd orders
        'objs' selects the NUV 2nd and FUV 3rd orders
            (higher resolution, lower S/N)

    Example
    -------
    # from http://galex.stsci.edu/GR6/?page=downloadlist&tilenum=5697&type=coaddS
    $ wget http://galex.stsci.edu/data/GR6/pipe/01-vsn/05697-VAR_J000836p545700/g/01-main/0001-img/07-try/VAR_J000836p545700-xg-gsp.fits.gz
    $ gunzip VAR_J000836p545700-xg-gsp.fits.gz

    >>> spblock = read_galex('VAR_J000836p545700-xg-gsp.fits')
    >>> spavg = spblock.average() # average over all spectra of the source (naive)
    >>> spblock.plotter()
    >>> spavg.plotter()
    """

    ff = pyfits.open(fitsfilename)
    bintable = ff[1].data

    # wavelength in angstroms
    wavelength = bintable.zero + np.arange(len(bintable.disp)) * bintable.disp
    xaxis = pyspeckit.units.SpectroscopicAxis(wavelength,units='angstroms')

    splist = [pyspeckit.Spectrum(data=bintable[orderselection][:,ii], error=bintable[orderselection+"err"][:,ii], xarr=xaxis, header=ff[1].header)
            for ii in xrange(bintable[orderselection].shape[1])]

    galexblock = pyspeckit.ObsBlock(splist)

    return galexblock



from matplotlib.backend_bases import NavigationToolbar2

class NavigationToolbar3(NavigationToolbar2):

    def press_pan(self, event):
        print "pan pressed"
        super(NavigationToolbar3,self).press_pan(self,event)

class MyNavToolbar(NavigationToolbar2):
    """wx/mpl NavToolbar hack with an additional tools user interaction.
    This class is necessary because simply adding a new togglable tool to the
    toolbar won't (1) radio-toggle between the new tool and the pan/zoom tools.
    (2) disable the pan/zoom tool modes in the associated subplot(s).
    """
    def __init__(self, canvas):
        super(MyNavToolbar, self).__init__(canvas)
        self.pan_tool  = self.FindById(self._NTB2_PAN)
        self.zoom_tool = self.FindById(self._NTB2_ZOOM)
        self.Bind(wx.EVT_TOOL, self.on_toggle_pan_zoom, self.zoom_tool)
        self.Bind(wx.EVT_TOOL, self.on_toggle_pan_zoom, self.pan_tool)

        self.user_tools = {}   # user_tools['tool_mode'] : wx.ToolBarToolBase

        self.InsertSeparator(5)
        self.add_user_tool('lasso', 6, icons.lasso_tool.ConvertToBitmap(), True, 'Lasso')
        self.add_user_tool('gate', 7, icons.gate_tool.ConvertToBitmap(), True, 'Gate')

    def add_user_tool(self, mode, pos, bmp, istoggle=True, shortHelp=''):
        """Adds a new user-defined tool to the toolbar.
        mode -- the value that MyNavToolbar.get_mode() will return if this tool 
                is toggled on
        pos -- the position in the toolbar to add the icon
        bmp -- a wx.Bitmap of the icon to use in the toolbar
        isToggle -- whether or not the new tool toggles on/off with the other 
                    togglable tools
        shortHelp -- the tooltip shown to the user for the new tool
        """
        tool_id = wx.NewId()
        self.user_tools[mode] = self.InsertSimpleTool(pos, tool_id, bmp,
                            isToggle=istoggle, shortHelpString=shortHelp)
        self.Bind(wx.EVT_TOOL, self.on_toggle_user_tool, self.user_tools[mode])

    def get_mode(self):
        """Use this rather than navtoolbar.mode
        """
        for mode, tool in self.user_tools.items():
            if tool.IsToggled():
                return mode
        return self.mode

    def untoggle_mpl_tools(self):
        """Hack city: Since I can't figure out how to change the way the 
        associated subplot(s) handles mouse events: I generate events to turn
        off whichever tool mode is enabled (if any). 
        This function needs to be called whenever any user-defined tool 
        (eg: lasso) is clicked.
        """
        if self.pan_tool.IsToggled():
            wx.PostEvent(
                self.GetEventHandler(), 
                wx.CommandEvent(wx.EVT_TOOL.typeId, self._NTB2_PAN)
            )
            self.ToggleTool(self._NTB2_PAN, False)
        elif self.zoom_tool.IsToggled():
            wx.PostEvent(
                self.GetEventHandler(),
                wx.CommandEvent(wx.EVT_TOOL.typeId, self._NTB2_ZOOM)
            )
            self.ToggleTool(self._NTB2_ZOOM, False)

    def on_toggle_user_tool(self, evt):
        """user tool click handler.
        """
        if evt.Checked():
            self.untoggle_mpl_tools()
            #untoggle other user tools
            for tool in self.user_tools.values():
                if tool.Id != evt.Id:
                    self.ToggleTool(tool.Id, False)

    def on_toggle_pan_zoom(self, evt):
        """Called when pan or zoom is toggled. 
        We need to manually untoggle user-defined tools.
        """
        if evt.Checked():
            for tool in self.user_tools.values():
                self.ToggleTool(tool.Id, False)
        # Make sure the regular pan/zoom handlers get the event
        evt.Skip()

    def reset_history(self):
        """More hacky junk to clear/reset the toolbar history.
        """
        self._views.clear()
        self._positions.clear()
        self.push_current()




"""
This wrapper is intended to replicate the inputs of agpy's showspec.splat_1d code.
It is uglier than the spectrum class.
"""

import pyspeckit

def splat_1d(filename=None,vmin=None,vmax=None,button=None,dobaseline=False,
        exclude=None,smooth=None,order=1,savepre=None,vcrop=True,
        vconv=None,vpars=None,spec=None,xtora=None,ytodec=None,
        specname=None,quiet=True,specnum=0,errspecnum=None,wcstype='',
        offset=0.0, continuum=0.0, annotatebaseline=False, plotspectrum=True,
        smoothto=None, xunits=None, units=None, conversion_factor=None,
        smoothtype='gaussian',convmode='same',maskspecnum=None,
        fignum=1, axis=None, autorefresh=False, title=None, color=None,
        label=None,clear=False, negamp=None, plotscale=1.0, voff=0.0, **kwargs):

    sp = pyspeckit.Spectrum(filename, specnum=specnum, errspecnum=errspecnum, wcstype=wcstype, **kwargs)
    sp.plotter.xmin = vmin
    sp.plotter.xmax = vmax
    sp.plotter.offset = offset
    if type(continuum) is str:
        sp.plotter.offset += sp.header.get(continuum)
    else:
        sp.plotter.offset += continuum
    if vcrop and vmin and vmax:
        sp.crop(vmin,vmax)

    if button != None:
        print "button keyword doesn't do anything"
    if maskspecnum:
        print "maskspecnum doesn't do anything [not implemented]"

    if specname:
        sp.plotter.title = specname
    if title:
        sp.plotter.title = title

    if smoothto:
        smooth = abs(smoothto/sp.xarr.cdelt())
    if smooth:
        sp.smooth(smooth, smoothtype=smoothtype, convmode=convmode)

    if xunits:
        sp.xarr.convert_to_unit(xunits)
    if units:
        sp.units = units

    if dobaseline:
        sp.baseline(order=order, exclude=exclude, annotate=annotatebaseline)

    if plotspectrum:
        if color is None:
            sp.plotter(figure=fignum, axis=axis, autorefresh=autorefresh, clear=clear, silent=quiet, reset_ylimits=True, title=title, plotscale=plotscale, xoffset=voff, label=label)
        else:
            sp.plotter(figure=fignum, axis=axis, autorefresh=autorefresh, color=color, clear=clear, silent=quiet, reset_ylimits=True, title=title, plotscale=plotscale, xoffset=voff, label=label)

    if savepre is not None:
        glon,glat = sp.header.get("GLON"),sp.header.get("GLAT")
        if glon is None or glat is None:
            raise ValueError("Header doesn't contain glon/glat.  Default savepre will not work.")
        if glat < 0: pm="" 
        else: pm = "+"
        savename = savepre + "G%07.3f%0s%07.3f_" % (glon,pm,glat) + sp.header.get('MOLECULE').replace(' ','') + sp.header.get('TRANSITI').replace(' ','')
        sp.plotter.savefig(savename+'.png')

    sp.fitspec = sp.specfit
    sp.spectrum = sp.data
    sp.axis = sp.plotter.axis
    sp.dv = sp.xarr.cdelt()
    sp.vind = sp.xarr
    sp.cube = sp.data
    sp.save = sp.write

    return sp



"""
A collection of wrappers
"""
from showspec_splat1d import splat_1d
from fit_gaussians_to_simple_spectra import fit_gaussians_to_simple_spectra
import fitnh3
import fith2co
from cube_fit import cube_fit
from load_IRAF_multispec import load_IRAF_multispec



"""
Storage for optical spectral line information.
"""

import numpy as np

# Format: name, units, vacuum?, display name
lines = {
         "H_alpha": [6564.614, 'Angstrom', True, r'$\mathrm{H}\alpha$'], \
         "H_beta":  [4862.721, 'Angstrom', True, r'$\mathrm{H}\beta$'], \
         "OIIIa":   [4960.295, 'Angstrom', True, r'$[\mathrm{OIII}]\lambda 4959\AA$'], \
         "OIIIb":   [5008.239, 'Angstrom', True, r'$[\mathrm{OIII}]\lambda 5007\AA$'], \
         "NIIa":    [6549.860, 'Angstrom', True, r'$[\mathrm{NII}]\lambda 6549\AA$'], \
         "NIIb":    [6585.270, 'Angstrom', True, r'$[\mathrm{NII}]\lambda 6585\AA$'], \
         "SIIa":    [6718.290, 'Angstrom', True, r'$[\mathrm{SII}]\lambda 6718\AA$'], \
         "SIIb":    [6732.680, 'Angstrom', True, r'$[\mathrm{SII}]\lambda 6732\AA$'], \
         "OI":      [6300.304, 'Angstrom', True, r'$[\mathrm{OI}]\lambda 6300\AA$'], \
         "OII":     [3727.319, 'Angstrom', True, r'$[\mathrm{OII}]\lambda 3727\AA$'], \
         "NeIII":   [3868.760, 'Angstrom', True, r'$[\mathrm{OII}]\lambda 3869\AA$']
}

xarr = []
for key in lines.keys(): 
    xarr.append(lines[key][0])
xarr = np.array(xarr)

indx = np.argsort(xarr)
xarr = np.sort(xarr)

name = []
for i, key in enumerate(lines.keys()): 
    name.append(lines.keys()[indx[i]])
name = np.array(name)

xunits = []
xvac = []
dname = []
for i, nombre in enumerate(name): 
    xunits.append(lines[nombre][1])
    xvac.append(lines[nombre][2])
    dname.append(lines[nombre][3])

xunits = np.array(xunits)
xvac = np.array(xvac)
dname = np.array(dname)

optical_lines = {'name': name, 'xarr': xarr, 'xunits': xunits, 'xvac': xvac, 'dname': dname}

def hydrogen(nu,nl, vacuum=True):
    """
    Compute the rest wavelength of Hydrogen recombination lines in angstroms
    """
    rydberg = 10973731.6 # m^-1
    protontoelectron = 1836.15266 # ratio

    lvac = 1.0/rydberg * 1./(1/float(nl)**2 - 1/float(nu)**2) * 1e10 * (1.0+1.0/protontoelectron)

    if not vacuum:
        import ref_index
        return ref_index.vac2air(lvac/10)*10
    else:
        return lvac



import pyfits
import pyspeckit


def load_IRAF_multispec(fitsfilename):
    """
    Load an IRAF multispec file as a list of spectra
    (a list of spectra can be passed to `pyspeckit.ObsBlock` if they are all of
    the same wavelength regime, but different objects or different spectra of
    the same object.  They can be passed to `pyspeckit.Spectra` if, e.g., you
    have an Echelle spectrum 
    """

    fitsfile = pyfits.open(fitsfilename)

    data, err, xax, header = pyspeckit.spectrum.readers.fits_readers.read_echelle(fitsfile[0])

    speclist = [pyspeckit.Spectrum(data=data[ii,:], error=err[ii,:], xarr=xax[ii,:], header=header) for ii in xrange(data.shape[0])]

    return speclist



import os
import numpy as np
from . import Writer

fitscheck = True
try: import pyfits
except ImportError: fitscheck = False

class write_fits(Writer):
    def write_data(self, filename = None, newsuffix = 'out', clobber = True, tolerance=1e-8,
            write_error=True, **kwargs ):
        """
        Write spectrum to fits file.
        """
        
        if not fitscheck: 
            print "Cannot write to fits - pyfits import failed."
            return
        
        if filename is None: fn = "{0}_{1}.fits".format(self.Spectrum.fileprefix, newsuffix)
        else: fn = filename

        header = self.Spectrum.header

        if header.get('CD1_1'): del header['CD1_1']

        # Generate a WCS header from the X-array
        if self.Spectrum.xarr._make_header(tolerance=tolerance):
            for k,v in self.Spectrum.xarr.wcshead.iteritems():
                if v is not None:
                    header.update(k,v)
            if write_error:
                data = np.array( [self.Spectrum.data, self.Spectrum.error] )
            else:
                data = self.Spectrum.data
            print "Writing a FITS-standard (linear-x-axis) spectrum to %s" % (fn)
        else:
            # if no header, overwrite header parameters that would be deceptive
            for k,v in self.Spectrum.xarr.wcshead.iteritems():
                if v is None:
                    if header.get(k): del header[k]
                else:
                    header.update(k,v)
            if write_error:
                data = np.array( [self.Spectrum.xarr, self.Spectrum.data, self.Spectrum.error] )
            else:
                data = np.array( [self.Spectrum.xarr, self.Spectrum.data] )
            print "Writing a nonlinear X-axis spectrum to %s (header keywords are not FITS-compliant)" % (fn)

        try:
            HDU = pyfits.PrimaryHDU(data=data, header=header)
        except AttributeError:
            print "Strange header error. Attempting workaround."
            HDU = pyfits.PrimaryHDU(data=data, header=pyfits.Header([pyfits.card.Card(k,v)  for k,v in header.iteritems()]))
        
        HDU.verify('fix')
        HDU.writeto(fn, clobber=clobber, output_verify='fix', **kwargs)



"""
==========================
PySpecKit ASCII Reader
==========================

Routines for reading in ASCII format spectra.  If atpy is not installed,
will use a very simple routine for reading in the data.  

.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
.. moduleauthor:: Jordan Mirocha <mirochaj@gmail.com>
"""
try:
    import atpy
    atpyOK = True
except ImportError:
    atpyOK = False
from .. import units
import numpy as np
from pyspeckit.specwarnings import warn
import readcol

def open_1d_txt(filename, xaxcol=0, datacol=1, errorcol=2, 
        text_reader='simple', atpytype='ascii', **kwargs):
    """
    Attempt to read a 1D spectrum from a text file assuming wavelength as the
    first column, data as the second, and (optionally) error as the third.  

    Reading can be done either with atpy or a 'simple' reader.  If you have an
    IPAC, CDS, or formally formatted table, you'll want to use atpy.

    If you have a simply formatted file of the form, e.g.
    # name name
    # unit unit
    data data
    data data

    kwargs are passed to atpy.Table
    """
    if text_reader in ('simple','readcol') or not atpyOK:
        if not atpyOK:
            warn("WARNING: atpy not installed; will use simple reader instead.")
        
        if text_reader == 'simple':
            data, error, XAxis, T = simple_txt(filename, xaxcol = xaxcol, 
                datacol = datacol, errorcol = errorcol, **kwargs)
        elif text_reader == 'readcol':
            Tlist = readcol.readcol(filename, twod=False, **kwargs)
            XAxis = units.SpectroscopicAxis(Tlist[xaxcol])
            data = Tlist[datacol]
            error = Tlist[errorcol]

            T = dummy_class()
            Tdict = readcol.readcol(filename, asDict=True, **kwargs)

            T.data = dummy_class()
            T.data.dtype = dummy_class()
            T.data.dtype.names = hdr
            T.columns = {}
            T.columns[T.data.dtype.names[xaxcol]] = dummy_class()
            T.columns[T.data.dtype.names[xaxcol]].unit = colunits[xaxcol]
            T.columns[T.data.dtype.names[datacol]] = dummy_class()
            T.columns[T.data.dtype.names[datacol]].unit = colunits[datacol]

        
    elif text_reader in ('atpy','asciitable'):
        T = atpy.Table(filename, type=atpytype, masked=True, **kwargs)
        
        xarr = T.data[T.data.dtype.names[xaxcol]]
        data = T.data[T.data.dtype.names[datacol]]
        if len(T.columns) > errorcol:
            error = T.data[T.data.dtype.names[errorcol]]
        else:
            # assume uniform, zero error
            error = data*0 

        if 'xunits' in T.keywords:
            xunits = T.keywords['xunits']
        else:
            xunits = 'unknown'

        XAxis = units.SpectroscopicAxis(xarr,xunits)
    
    # Need this in Spectrum class to correctly parse header    
    T.xaxcol = xaxcol
    T.datacol = datacol 

    return data, error, XAxis, T

def simple_txt(filename, xaxcol=0, datacol=1, errorcol=2, skiplines=0, **kwargs):
    """
    Very simple method for reading columns from ASCII file.
    """
    
    f = open(filename, 'r')
    
    hdr = None
    colunits = []
    coldata = []
    for ii, line in enumerate(f):
        
        # Ignore blank lines
        if not line.strip(): 
            continue
        
        # Possibly read in header
        if line.split()[0][0] == '#': 
            if (ii) == (0+skiplines): 
                hdr = line[1:].split()
            if (ii) == (1+skiplines): 
                colunits = line[1:].split()
            
            continue
        
        coldata.append(line.split())
        
        for j, element in enumerate(coldata[-1]):
            try: 
                coldata[-1][j] = float(element)
            except ValueError: 
                coldata[-1][j] = str(element)
                
    f.close()
    
    if not colunits:
        colunits = ['unknown', 'unknown', 'unknown']
    if not hdr:
        hdr = ['unknown', 'unknown', 'unknown']    
        
    N = len(hdr)
    coldata = zip(*coldata)     
            
    # Prepare to return data
    data = coldata[datacol]
    xarr = coldata[xaxcol]
    if errorcol > N - 1:
        error = np.array(data)*0
    else:
        error = coldata[errorcol]

    if len(error) != len(data):
        raise ValueError("Data and Error lengths do not match.")

    XAxis = units.SpectroscopicAxis(xarr, colunits[xaxcol]) 
    
    # Create atPy style Table instance
    T = dummy_class()
    T.data = dummy_class()
    T.data.dtype = dummy_class()
    T.data.dtype.names = hdr
    T.columns = {}
    T.columns[T.data.dtype.names[xaxcol]] = dummy_class()
    T.columns[T.data.dtype.names[xaxcol]].unit = colunits[xaxcol]
    T.columns[T.data.dtype.names[datacol]] = dummy_class()
    T.columns[T.data.dtype.names[datacol]].unit = colunits[datacol]
                
    return np.array(data), np.array(error), XAxis, T
    
class dummy_class:
    def __init__(self):
        pass





"""
Override components of python's builting warnings (as suggested by the manual)
"""

import warnings
import sys

def showwarning(message, category, filename, lineno, file=None, line=None):
    """Hook to write a warning to a file; replace if you like."""
    if file is None:
        file = sys.stderr
    try:
        file.write(message.message+"\n")
        #file.write(formatwarning(message, category, filename, lineno, line))
    except IOError:
        pass # the file (probably stderr) is invalid - this warning gets lost.

# This is not overwritten, but may be in the future... kept here so we know what to edit
#def formatwarning(message, category, filename, lineno, line=None):
#    """Function to format a warning the standard way."""
#    s =  "%s:%s: %s: %s\n" % (filename, lineno, category.__name__, message)
#    line = linecache.getline(filename, lineno) if line is None else line
#    if line:
#        line = line.strip()
#        s += "  %s\n" % line
#    return s

warnings.showwarning = showwarning

warn = warnings.warn



"""
A flexible Spectral Model class.  Allows you to add and subtract, fix and unfix functions, etc.

Goal is to allow adding/subtracting/multiplying components...
(like these: https://astrophysics.gsfc.nasa.gov/XSPECwiki/Basic_single_dataset)
"""



class ModelComponent(object):
    def __init__(self, function, name, parameters=None, fixed=False):
        """
        Each Model object consists of a function...
        """
        self.function = function
        self.name = name
        self.parameters = parameters
        self.fixed = fixed

    def __call__(self, xaxis):
        """
        A model is defined over an X-axis.  It can do anything here, but must
        return an array of length len(xaxis).
        """

        return sum_of_model_functions

class AdditiveModelComponent(ModelComponent):
    pass
    
class MultiplicativeModelComponent(ModelComponent):
    pass
    

class Model(ModelComponent):
    """
    A model will consist of a sum of its children?

    """
    def __init__(self):
        """
        Generic model...
        """
        self.additive_components = []
        self.multiplicative_components = []

    def __call__(self):
        """
        Return sum/
        """

    def __add__(self, other):
        """
        Models can be added together
        """
        self.additive_components.append(other)

        return self

    def __sub__(self, other):
        """
        Models can be removed
        """
        try:
            self.additive_components.remove(other)
        except ValueError:
            print "Can only remove components that have been included in the model"

        return self

    def __mul__(self, other):
        """
        Models can be added multiplicatively too
        """
        self.multiplicative_components.append(other)
        return self

    def __div__(self, other):
        """
        And models can be divided out
        """
        try:
            self.multiplicative_components.remove(other)
        except ValueError:
            print "Can only remove components that have been included in the model"

        return self



"""
~~~~~~~~
cubes.py
~~~~~~~~

From `agpy <http://code.google.com/p/agpy/source/browse/trunk/agpy/cubes.py>`_,
contains functions to perform various transformations on data cubes and their
headers.  

########

"""
from numpy import sqrt,repeat,indices,newaxis,pi,cos,sin,array,mean,sum,nansum
from math import acos,atan2,tan
import numpy
import copy
import os
import tempfile
import posang # agpy code
import pyspeckit
from pyspeckit.specwarnings import warn
try:
    from AG_fft_tools.convolve import smooth
    from pyspeckit.parallel_map import parallel_map
    smoothOK = True
except ImportError:
    smoothOK = False
try:
    import coords
    coordsOK = True
except ImportError:
    coordsOK = False
try: # pywcs + pyfits OR astropy, no mixing!
    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs
    wcsOK = True
except ImportError:
    import pyfits
    try:
        import pywcs
        wcsOK = True
    except ImportError:
        wcsOK = False

dtor = pi/180.0

def flatten_header(header):
    """
    Attempt to turn an N-dimensional fits header into a 2-dimensional header
    Turns all CRPIX[>2] etc. into new keywords with suffix 'A'

    header must be a pyfits.Header instance
    """

    if not isinstance(header,pyfits.Header):
        raise Exception("flatten_header requires a pyfits.Header instance")

    newheader = header.copy()

    for key in newheader.keys():
        try:
            if int(key[-1]) >= 3 and key[:2] in ['CD','CR','CT','CU','NA']:
                newheader.rename_key(key,'A'+key,force=True)
        except ValueError:
            # if key[-1] is not an int
            pass
        except IndexError:
            # if len(key) < 2
            pass
    newheader.update('NAXIS',2)

    return newheader

def speccen_header(header,lon=None,lat=None):
    """
    Turn a cube header into a spectrum header, retaining RA/Dec vals where possible
    (speccen is like flatten; spec-ify would be better but, specify?  nah)

    Assumes 3rd axis is velocity
    """
    newheader = header.copy()
    newheader.update('CRVAL1',header.get('CRVAL3'))
    newheader.update('CRPIX1',header.get('CRPIX3'))
    if 'CD1_1' in header: newheader.rename_key('CD1_1','OLDCD1_1')
    elif 'CDELT1' in header: newheader.rename_key('CDELT1','OLDCDEL1')
    if 'CD3_3' in header: newheader.update('CDELT1',header.get('CD3_3'))
    elif 'CDELT3' in header: newheader.update('CDELT1',header.get('CDELT3'))
    newheader.update('CTYPE1','VRAD')
    if header.get('CUNIT3'): newheader.update('CUNIT1',header.get('CUNIT3'))
    else: 
        print "Assuming CUNIT3 is km/s in speccen_header"
        newheader.update('CUNIT1','km/s')
    newheader.update('CRPIX2',1)
    newheader.update('CTYPE2','RA---TAN')
    newheader.update('CRPIX3',1)
    newheader.update('CTYPE3','DEC--TAN')

    if lon is not None: newheader.update('CRVAL2',lon)
    if lat is not None: newheader.update('CRVAL3',lat)

    if 'CD2_2' in header: newheader.rename_key('CD2_2','OLDCD2_2')
    if 'CD3_3' in header: newheader.rename_key('CD3_3','OLDCD3_3')

    return newheader

def extract_aperture(cube,ap,r_mask=False,wcs=None,coordsys='galactic',wunit='arcsec',debug=False):
    """
    Extract an aperture from a data cube.  E.g. to acquire a spectrum
    of an outflow that is extended.

    Cube should have shape [z,y,x], e.g. 
    cube = pyfits.getdata('datacube.fits')

    Apertures are specified in PIXEL units with an origin of 0,0 (NOT the 1,1
    fits standard!) unless wcs and coordsys are specified
    
    INPUTS:
        wcs - a pywcs.WCS instance associated with the data cube
        coordsys - the coordinate system the aperture is specified in.
            Options are 'celestial' and 'galactic'.  Default is 'galactic'
        wunit - units of width/height.  default 'arcsec', options 'arcmin' and 'degree'

    For a circular aperture, len(ap)=3:
        ap = [xcen,ycen,radius]

    For an elliptical aperture, len(ap)=5:
        ap = [xcen,ycen,height,width,PA]

    Optional inputs:
        r_mask - return mask in addition to spectrum (for error checking?)
    """

    if wcs is not None and coordsys is not None:
        if debug: print "Converting aperture ",ap,
        ap = aper_world2pix(ap,wcs,coordsys=coordsys,wunit=wunit)
        if debug: print " to ",ap

    if len(ap) == 3:
        sh = cube.shape
        yind,xind = indices(sh[1:3]) # recall that python indices are backwards
        dis = sqrt((xind-ap[0])**2+(yind-ap[1])**2)
        mask = dis < ap[2]
    elif len(ap) == 5:
        yinds,xinds = indices(cube.shape[1:3])
        th = (ap[4])*dtor
        xindr = (xinds-ap[0])*cos(th)  + (yinds-ap[1])*sin(th)
        yindr = (xinds-ap[0])*-sin(th) + (yinds-ap[1])*cos(th)
        ratio = max(ap[2:4])/min(ap[2:4])
        mask = sqrt( (xindr*ratio)**2 + yindr**2) < max(ap[2:4])
    else:
        raise Exception("Wrong number of parameters.  Need either 3 parameters "+
                "for a circular aperture or 5 parameters for an elliptical "+ 
                "aperture.")

    npixinmask = mask.sum()
    mask3d = repeat(mask[newaxis,:,:],cube.shape[0],axis=0)
    spec = nansum(nansum((cube*mask3d),axis=2),axis=1) / npixinmask

    if r_mask:
        return spec,mask
    else:
        return spec

def integ(file,vrange,xcen=None,xwidth=None,ycen=None,ywidth=None,**kwargs):
    """
    wrapper of subimage_integ that defaults to using the full image
    """
    if isinstance(file,pyfits.PrimaryHDU):
        header = file.header
        cube = file.data
    elif isinstance(file,pyfits.HDUList):
        header = file[0].header
        cube = file[0].data
    else:
        file = pyfits.open(file)
        header = file[0].header
        cube = file[0].data

    if None in [xcen,xwidth,ycen,ywidth]:
        xcen = header['NAXIS1'] / 2
        xwidth = xcen
        ycen = header['NAXIS2'] / 2
        ywidth = ycen

    return subimage_integ(cube,xcen,xwidth,ycen,ywidth,vrange,header=header,**kwargs)

def subimage_integ(cube, xcen, xwidth, ycen, ywidth, vrange, header=None,
        average=mean, dvmult=False, return_HDU=False, units="pixels",
        zunits=None):
    """
    Returns a sub-image from a data cube integrated over the specified velocity range

    All units assumed to be pixel units

    cube has dimensions (velocity, y, x)

    xwidth and ywidth are "radius" values, i.e. half the length that will be extracted

    if dvmult is set, multiple the average by DV (this is useful if you set
    average=sum and dvmul=True to get an integrated value)

    """

    if header:
        flathead = flatten_header(header.copy())
        wcs = pywcs.WCS(header=flathead)
        if header.get('CD3_3'): CD3 = header.get('CD3_3')
        else: CD3 = header.get('CDELT3')

    if units=="pixels":
        xlo = int( max([xcen-xwidth,0])              )
        ylo = int( max([ycen-ywidth,0])              )
        xhi = int( min([xcen+xwidth,cube.shape[2]])  )
        yhi = int( min([ycen+ywidth,cube.shape[1]])  )
    elif units=="wcs" and header:
        newxcen,newycen = wcs.wcs_world2pix(xcen,ycen,0)
        try:
            newxwid,newywid = xwidth / abs(wcs.wcs.cd[0,0]), ywidth / abs(wcs.wcs.cd[1,1])
        except AttributeError:
            newxwid,newywid = xwidth / abs(wcs.wcs.cdelt[0]), ywidth / abs(wcs.wcs.cdelt[1])
        xlo = int( max([newxcen-newxwid,0]) )
        ylo = int( max([newycen-newywid,0]) )
        xhi = int( min([newxcen+newxwid,cube.shape[2]]) )
        yhi = int( min([newycen+newywid,cube.shape[1]]) )
    else:
        print "Can only use wcs if you pass a header."

    if zunits is None:
        zunits = units
    if zunits == 'pixels':
        zrange = vrange
    if zunits == 'wcs':
        zrange = ( array(vrange)-header.get('CRVAL3') ) / CD3 - 1 + header.get('CRPIX3')

    subim = average(cube[zrange[0]:zrange[1],ylo:yhi,xlo:xhi],axis=0)
    if dvmult and CD3: subim *= CD3
    elif dvmult: print "Error: could not multiply by dv; CD3=",CD3

    if header is None:
        return subim
    else:
        crv1,crv2 = wcs.wcs_pix2world(xlo,ylo,0)

        flathead['CRVAL1'] = crv1[0]
        flathead['CRVAL2'] = crv2[0]
        flathead['CRPIX1'] = 1
        flathead['CRPIX2'] = 1
        
        if return_HDU:
            return pyfits.PrimaryHDU(data=subim,header=flathead)
        else:
            return subim,flathead

def subcube(cube, xcen, xwidth, ycen, ywidth, header=None,
        dvmult=False, return_HDU=False, units="pixels",
        widthunits="pixels"):
    """
    Crops a data cube

    All units assumed to be pixel units

    cube has dimensions (velocity, y, x)

    xwidth and ywidth are "radius" values, i.e. half the length that will be extracted

    if dvmult is set, multiple the average by DV (this is useful if you set
    average=sum and dvmul=True to get an integrated value)

    """

    if header:
        newheader = header.copy()
        flathead = flatten_header(header.copy())
        wcs = pywcs.WCS(header=flathead)

    if widthunits == "pixels":
        newxwid, newywid = xwidth, ywidth
    elif widthunits == "wcs":
        try:
            newxwid,newywid = xwidth / abs(wcs.wcs.cd[0,0]), ywidth / abs(wcs.wcs.cd[1,1])
        except AttributeError:
            newxwid,newywid = xwidth / abs(wcs.wcs.cdelt[0]), ywidth / abs(wcs.wcs.cdelt[1])
    else:
        raise Exception("widthunits must be either 'wcs' or 'pixels'")

    if units=="pixels":
        newxcen,newycen = xcen,ycen
    elif units=="wcs" and header:
        newxcen,newycen = wcs.wcs_world2pix(xcen,ycen,0)
    else:
        raise Exception("units must be either 'wcs' or 'pixels'")

    x1 = int( numpy.floor( max([newxcen-newxwid,0]) ) )
    y1 = int( numpy.floor( max([newycen-newywid,0]) ) )
    x2 = int( numpy.ceil( min([newxcen+newxwid,cube.shape[2]]) ) )
    y2 = int( numpy.ceil( min([newycen+newywid,cube.shape[1]]) ) )

    xhi = max(x1,x2)
    xlo = min(x1,x2)
    yhi = max(y1,y2)
    ylo = min(y1,y2)

    subim = cube[:,ylo:yhi,xlo:xhi]

    if return_HDU:

        xmid_sky,ymid_sky = wcs.wcs_pix2world(xlo+xwidth,ylo+ywidth,0)

        newheader.update('CRVAL1',xmid_sky[0])
        newheader.update('CRVAL2',ymid_sky[0])
        newheader.update('CRPIX1',1+xwidth)
        newheader.update('CRPIX2',1+ywidth)
        
        newHDU =  pyfits.PrimaryHDU(data=subim,header=newheader)
        if newHDU.header.get('NAXIS1') == 0 or newHDU.header.get('NAXIS2') == 0:
            raise Exception("Cube has been cropped to 0 in one dimension")

        return newHDU
    else:
        return subim

def aper_world2pix(ap,wcs,coordsys='galactic',wunit='arcsec'):
    """
    Converts an elliptical aperture (x,y,width,height,PA) from
    WCS to pixel coordinates given an input wcs (an instance
    of the pywcs.WCS class).  Must be a 2D WCS header.


    """
    if not coordsOK:
        raise ImportError( "cubes.py requires coords for aper_world2pix and coords_in_image" )

    convopt = {'arcsec':3600.0,'arcmin':60.0,'degree':1.0}
    try:
        conv = convopt[wunit]
    except:
        raise Exception("Must specify wunit='arcsec','arcmin', or 'degree'")

    if len(wcs.wcs.cdelt) != 2:
        raise Exception("WCS header is not strictly 2-dimensional.  Look for 3D keywords.")
    pos = coords.Position((ap[0],ap[1]),system=coordsys)
    if wcs.wcs.ctype[0][:2] == 'RA':
        ra,dec = pos.j2000()
        corrfactor = cos(dec*dtor)
    elif wcs.wcs.ctype[0][:4] == 'GLON':
        ra,dec = pos.galactic()
        corrfactor=1
    # workaround for a broken wcs.wcs_world2pix
    try:
        radif = (wcs.wcs.crval[0]-ra)*dtor
        gamma = acos(cos(dec*dtor)*cos(wcs.wcs.crval[1]*dtor)*cos(radif)+sin(dec*dtor)*sin(wcs.wcs.crval[1]*dtor)) / dtor
        theta = atan2( sin(radif) , ( tan(dec*dtor)*cos(wcs.wcs.crval[1]*dtor)-sin(wcs.wcs.crval[1]*dtor)*cos(radif) ) )
        x = -gamma * sin(theta) / wcs.wcs.cd[0,0] + wcs.wcs.crpix[0]
        y = gamma * cos(theta) / wcs.wcs.cd[1,1] + wcs.wcs.crpix[1]
    except:
        radif = (wcs.wcs.crval[0]-ra)*dtor
        gamma = acos(cos(dec*dtor)*cos(wcs.wcs.crval[1]*dtor)*cos(radif)+sin(dec*dtor)*sin(wcs.wcs.crval[1]*dtor)) / dtor
        theta = atan2( sin(radif) , ( tan(dec*dtor)*cos(wcs.wcs.crval[1]*dtor)-sin(wcs.wcs.crval[1]*dtor)*cos(radif) ) )
        x = -gamma * sin(theta) / wcs.wcs.cdelt[0] + wcs.wcs.crpix[0]
        y = gamma * cos(theta) / wcs.wcs.cdelt[1] + wcs.wcs.crpix[1]
    
    #print "DEBUG: x,y from math (vectors): ",x,y
    #x,y = wcs.wcs_world2pix(ra,dec,0)  # convert WCS coordinate to pixel coordinate (0 is origin, do not use fits convention)
    #print "DEBUG: x,y from wcs: ",x,y
    try:
        x=x[0] - 1 # change from FITS to python convention
        y=y[0] - 1 # change from FITS to python convention
        #print "DEBUG: x,y from math: ",x,y
    except:
        pass
    # cd is default, cdelt is backup
    if len(ap) == 5:
        try:
            width  = ap[2] / conv / abs(wcs.wcs.cd[0,0])  # first is width, second is height in DS9 PA convention
            height = ap[3] / conv / abs(wcs.wcs.cd[0,0])
        except:
            width  = ap[2] / conv / abs(wcs.wcs.cdelt[0])  # first is width, second is height in DS9 PA convention
            height = ap[3] / conv / abs(wcs.wcs.cdelt[0])
        PA = ap[4] 
        apold = copy.copy(ap)
        ap = [x,y,width,height,PA]
    elif len(ap) == 3:
        try:
            width  = ap[2] / conv / abs(wcs.wcs.cd[0,0])  # first is width, second is height in DS9 PA convention
        except:
            width  = ap[2] / conv / abs(wcs.wcs.cdelt[0])  # first is width, second is height in DS9 PA convention
        apold = copy.copy(ap)
        ap = [x,y,width]

    return ap


def getspec(lon,lat,rad,cube,header,r_fits=True,inherit=True,wunit='arcsec'):
    """
    Given a longitude, latitude, aperture radius (arcsec), and a cube file,
    return a .fits file or a spectrum.
    
    lon,lat - longitude and latitude center of a circular aperture in WCS coordinates
                must be in coordinate system of the file
    rad     - radius (default degrees) of aperture
    """
    if not wcsOK:
        raise ImportError( "cubes.py requires astropy.wcs or pywcs for some subimage_integ,aper_wordl2pix,getspec, and coords_in_image" )

    convopt = {'arcsec':1.0,'arcmin':60.0,'degree':3600.0}

    flathead = flatten_header(header)
    wcs = pywcs.WCS(flathead)
    if wcs.wcs.ctype[0][:2] == 'RA':
      coordsys='celestial'
    elif wcs.wcs.ctype[0][:4] == 'GLON':
      coordsys='galactic'
    spec = extract_aperture(cube,[lon,lat,rad],wcs=wcs,
            coordsys=coordsys,wunit=wunit)

    if nansum(spec) == 0:
        print "Total of extracted spectrum was zero. lon,lat,rad: ",lon,lat,rad #  Tracing to find your problem."
        #import pdb; pdb.set_trace()

    if r_fits:
        if inherit:
            newhead = header.copy()
        else:
            newhead = pyfits.Header()
        try:
            newhead.update('CD1_1',header['CD3_3'])
        except KeyError:
            newhead.update('CD1_1',header['CDELT3'])
        newhead.update('CRPIX1',header['CRPIX3'])
        newhead.update('CRVAL1',header['CRVAL3'])
        try:
            newhead.update('CTYPE1',header['CTYPE3'])
        except KeyError:
            newhead.update('CTYPE1',"VRAD")
        try:
            newhead.update('CUNIT1',header['CUNIT3'])
        except KeyError:
            print "Header did not contain CUNIT3 keyword.  Defaulting to km/s"
            newhead.update('CUNIT1',"km/s")
        newhead.update('BUNIT',header['BUNIT'])
        newhead.update('APGLON',lon)
        newhead.update('APGLAT',lat)
        newhead.update('APRAD',rad*convopt[wunit],comment='arcseconds') # radius in arcsec
        newfile = pyfits.PrimaryHDU(data=spec,header=newhead)
        return newfile
    else:
        return spec

def getspec_reg(cubefilename,region,**kwargs):
    """
    Aperture extraction from a cube using a pyregion circle region

    The region must be in the same coordinate system as the cube header

    .. warning:: The second argument of getspec_reg requires a pyregion region list, 
        and therefore this code depends on `pyregion <http://leejjoon.github.com/pyregion/>`_.
    """

    ds9tocoords = {'fk5':'celestial','galactic':'galactic','icrs':'celestial'}

    if region.name != 'circle':
        raise Exception("Only circular apertures are implemented so far")

    l,b,r = region.coord_list
    #pos = coords.Position([l,b],system=ds9tocoords[region.coord_format])
    if isinstance(cubefilename,pyfits.HDUList):
        cubefile = cubefilename
    else:
        cubefile = pyfits.open(cubefilename)
    header = cubefile[0].header
    cube = cubefile[0].data
    if len(cube.shape) == 4: cube = cube[0,:,:,:]

    sp = getspec(l,b,r,cube,header,wunit='degree',**kwargs)

    return sp

def coords_in_image(fitsfile,lon,lat,system='galactic'):
    """
    Determine whether the coordinates are inside the image
    """
    if not coordsOK:
        raise ImportError( "cubes.py requires coords for aper_world2pix and coords_in_image" )
    if not wcsOK:
        raise ImportError( "cubes.py requires astropy.wcs or pywcs for some subimage_integ,aper_wordl2pix,getspec, and coords_in_image" )
    if not isinstance(fitsfile,pyfits.HDUList):
        fitsfile = pyfits.open(fitsfile)

    wcs = pywcs.WCS(flatten_header(fitsfile[0].header))

    if 'RA' in wcs.wcs.ctype[0]:
        pos = coords.Position((lon,lat),system=system)
        lon,lat = pos.j2000()
    if 'GLON' in wcs.wcs.ctype[0]:
        pos = coords.Position((lon,lat),system=system)
        lon,lat = pos.galactic()

    x,y = wcs.wcs_world2pix(lon,lat,0)
    #DEBUG print x,y,wcs.naxis1,wcs.naxis2
    if (0 < x < wcs.naxis1) and (0 < y < wcs.naxis2):
        return True
    else:
        return False

def spectral_smooth(cube, smooth_factor, **kwargs):
    """
    Smooth the cube along the spectral direction
    """

    newcube = numpy.zeros(cube.shape)
    yy,xx = numpy.indices(cube.shape[1:])

    for (x,y) in zip(xx.flat,yy.flat):
        newcube[:,y,x] = pyspeckit.smooth.smooth(cube[:,y,x], smooth_factor, **kwargs)

    return newcube

def plane_smooth(cube,cubedim=0,parallel=True,numcores=None,**kwargs):
    """
    parallel-map the smooth function

    parallel - defaults True.  Set to false if you want serial (for debug
        purposes?)
    numcores - pass to parallel_map (None = use all available)
    """
    if not smoothOK:
        return

    if cubedim != 0:
        cube = cube.swapaxes(0,cubedim)

    cubelist = [cube[ii,:,:] for ii in xrange(cube.shape[0])]

    Psmooth = lambda C: smooth(C,**kwargs)

    if parallel:
        smoothcube = array(parallel_map(Psmooth,cubelist,numcores=numcores))
    else:
        smoothcube = array(map(Psmooth,cubelist))
    
    if cubedim != 0:
        smoothcube = smoothcube.swapaxes(0,cubedim)

    return smoothcube


try:
    import montage 

    def rotcrop_cube(x1, y1, x2, y2, cubename, outname, xwidth=25, ywidth=25,
            in_system='galactic',  out_system='equatorial', clobber=True):
        """
        Crop a data cube and then rotate it with montage

        """

        cubefile = pyfits.open(cubename)

        if not coordsOK:
            raise ImportError( "cubes.py requires coords for rotcrop_cube" )
        pos1 = coords.Position([x1,y1],system=in_system)
        pos2 = coords.Position([x2,y2],system=in_system)

        if cubefile[0].header.get('CTYPE1')[:2] == 'RA':
            x1,y1 = pos1.j2000()
            x2,y2 = pos2.j2000()
            coord_system = 'celestial'
        elif  cubefile[0].header.get('CTYPE1')[:2] == 'GLON':
            x1,y1 = pos1.galactic()
            x2,y2 = pos2.galactic()
            coord_system = 'galactic'

        xcen = (x1+x2)/2.0
        ycen = (y1+y2)/2.0

        sc = subcube(cubefile[0].data, xcen, xwidth, ycen, ywidth, 
                widthunits='pixels', units="wcs", header=cubefile[0].header,
                return_HDU=True)
        # note: there should be no security risk here because pyfits' writeto
        # will not overwrite by default
        tempcube = tempfile.mktemp(suffix='.fits')
        sc.writeto(tempcube)
        
        pa = posang.posang(x1,y1,x2,y2,system=coord_system) - 90

        newheader = sc.header.copy()
        cd11 = newheader.get('CDELT1') if newheader.get('CDELT1') else newheader.get('CD1_1')
        cd22 = newheader.get('CDELT2') if newheader.get('CDELT2') else newheader.get('CD2_2')
        cd12 = newheader.get('CD1_2') if newheader.get('CD1_2') else 0.0
        cd21 = newheader.get('CD2_1') if newheader.get('CD2_1') else 0.0
        cdelt = numpy.sqrt(cd11**2+cd12**2)

        tempheader = tempfile.mktemp(suffix='.hdr')
        ycensign = "+" if numpy.sign(ycen) >= 0 else "-"
        montage.mHdr("%s %1s%s" % (xcen, ycensign, numpy.abs(ycen)), xwidth*cdelt,
                tempheader, system=out_system, height=ywidth*cdelt,
                pix_size=cdelt*3600.0, rotation=pa)
        os.system("sed -i bck '/END/d' %s" % (tempheader))
        newheader2 = pyfits.Header()
        newheader2.fromTxtFile(tempheader)
        for key in ('CRPIX3','CRVAL3','CDELT3','CD3_3','CUNIT3','WCSTYPE3','CTYPE3'):
            if newheader.get(key):
                newheader2.update(key,newheader.get(key))
        if newheader.get('CD3_3') and newheader2.get('CDELT3') is None:
            newheader2.update('CDELT3',newheader.get('CD3_3'))
        newheader2.toTxtFile(tempheader,clobber=True)
        #if newheader2.get('CDELT3') is None:
        #    raise Exception("No CD3_3 or CDELT3 in header.")

        montage.wrappers.reproject_cube(tempcube,outname,header=tempheader,clobber=clobber)
        #print "\n",outname
        #os.system('imhead %s | grep CDELT' % outname)

        # AWFUL hack because montage removes CDELT3
        tempcube = pyfits.open(outname)
        tempcube.header = newheader2
        #if tempcube.header.get('CDELT3') is None:
        #    raise Exception("No CD3_3 or CDELT3 in header.")
        #print tempcube.header.get('CDELT3')
        tempcube.writeto(outname,clobber=True)
        #print tempcube.get('CDELT3')
        #print "\n",outname
        #os.system('imhead %s | grep CDELT' % outname)

        #print "\nnewheader2"
        #print newheader2.ascard
        #print
        
        return

except:
    pass



"""
======================================
Generalized hyperfine component fitter
======================================
.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
"""
import numpy as np
import model,fitter

ckms = 2.99792458e5

class hyperfinemodel(object):
    """
    Wrapper for the hyperfine model class.  Specify the offsets and relative
    strengths when initializing, then you've got yourself a hyperfine modeler.
    """

    def __init__(self, line_names, voff_lines_dict, freq_dict, line_strength_dict, relative_strength_total_degeneracy):
        """
        Initialize the various parameters defining the hyperfine transitions

        line_names is a LIST of the line names to be used as indices for the dictionaries

        voff_lines_dict is a linename:v_off dictionary of velocity offsets for the hyperfine components.  Technically,
            this is redundant with freq_dict
        freq_dict - frequencies of the indvidual transitions

        line_strength_dict - Relative strengths of the hyperfine components, usually determined by their degeneracy and 
            Einstein A coefficients
        """
        self.line_names = line_names
        self.voff_lines_dict = voff_lines_dict
        self.freq_dict = freq_dict
        self.line_strength_dict = line_strength_dict
        self.relative_strength_total_degeneracy = relative_strength_total_degeneracy

        self.fitter = model.SpectralModel(self,4,
            parnames=['Tex','tau','center','width'], 
            parlimited=[(False,False), (True,False), (False,False), (True,False)], 
            parlimits=[(0,0), (0,0), (0,0), (0,0)],
            shortvarnames=("T_{ex}","\\tau","v","\\sigma"), # specify the parameter names (TeX is OK)
            fitunits='Hz' )

        self.vheight_fitter = model.SpectralModel(fitter.vheightmodel(self),5,
            parnames=['height','Tex','tau','center','width'], 
            parlimited=[(False,False), (False,False), (True,False), (False,False), (True,False)], 
            parlimits=[(0,0), (0,0), (0,0), (0,0), (0,0)],
            shortvarnames=("H","T_{ex}","\\tau","v","\\sigma"), # specify the parameter names (TeX is OK)
            fitunits='Hz' )

        self.ampfitter = model.SpectralModel(self.hyperfine_amp,3,
            parnames=['amp','center','width'], 
            parlimited=[(False,False), (False,False), (True,False)], 
            parlimits=[(0,0), (0,0), (0,0)],
            shortvarnames=("amp","v","\\sigma"), # specify the parameter names (TeX is OK)
            fitunits='Hz' )

    def __call__(self, *args, **kwargs):
        """
        Generate a model spectrum given an excitation temperature, optical depth, offset velocity, and velocity width.
        """
        return self.hyperfine(*args,**kwargs)

    def hyperfine_amp(self, xarr, amp=None, xoff_v=0.0, width=1.0, 
            return_hyperfine_components=False, Tbackground=2.73, Tex=5.0, tau=0.1):
        """
        wrapper of self.hyperfine with order of arguments changed
        """ 
        return self.hyperfine(xarr, amp=amp, Tex=Tex, tau=tau, xoff_v=xoff_v,
                width=width, return_hyperfine_components=return_hyperfine_components,
                Tbackground=Tbackground)


    def hyperfine(self, xarr, Tex=5.0, tau=0.1, xoff_v=0.0, width=1.0, 
            return_hyperfine_components=False, Tbackground=2.73, amp=None ):
        """
        Generate a model spectrum given an excitation temperature, optical depth, offset velocity, and velocity width.
        """

        # Convert X-units to frequency in Hz
        xarr = xarr.as_unit('Hz')

        # Ensure parameters are scalar
        if not np.isscalar(Tex): Tex = Tex.squeeze()
        if not np.isscalar(tau): tau = tau.squeeze()
        if not np.isscalar(xoff_v): xoff_v = xoff_v.squeeze()
        if not np.isscalar(width): width = width.squeeze()

        # Generate an optical depth spectrum as a function of the X-axis 
        tau_nu_cumul = np.zeros(len(xarr))
        # Error check: inputing NANs results in meaningless output - return without computing a model
        if np.any(np.isnan((tau,Tex,width,xoff_v))):
            if return_hyperfine_components:
                return [tau_nu_cumul] * len(self.line_names)
            else:
                return tau_nu_cumul

        components =[]
        for linename in self.line_names:
            voff_lines = np.array(self.voff_lines_dict[linename])
      
            lines = (1-voff_lines/ckms)*self.freq_dict[linename]
            if width == 0:
                tau_nu = xarr*0
            else:
                nuwidth = np.abs(width/ckms*lines)
                nuoff = xoff_v/ckms*lines
                # the total optical depth, which is being fitted, should be the sum of the components
                tau_line = (tau * self.line_strength_dict[linename]/self.relative_strength_total_degeneracy[linename]) 
          
                tau_nu = np.array(tau_line * np.exp(-(xarr+nuoff-self.freq_dict[linename])**2/(2.0*nuwidth**2)))
                tau_nu[tau_nu!=tau_nu] = 0 # avoid nans
                components.append( tau_nu )
            tau_nu_cumul += tau_nu

        # add a list of the individual 'component' spectra to the total components...

        if return_hyperfine_components:
            if amp is None:
                return (1.0-np.exp(-np.array(components)))*(Tex-Tbackground)
            else:
                comps = (1.0-np.exp(-np.array(components)))*(Tex-Tbackground)
                return comps/comps.max() * amp

        spec = (1.0-np.exp(-np.array(tau_nu_cumul)))*(Tex-Tbackground)
      
        if amp is None:
            return spec
        else:
            return spec/spec.max() * amp



